import unittest
import json
from copy import deepcopy
from unittest import mock
from neo3 import contracts, storage
from neo3.core import cryptography, types, utils, serialization


class ContractGroupTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        private_key = b'\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
        cls.keypair = cryptography.KeyPair(private_key)

        # capture from C#
        cls.expected_json = {"pubkey":"033d523f36a732974c0f7dbdfafb5206ecd087211366a274190f05b86d357f4bad","signature":"AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=="}

    def test_is_valid(self):
        """
        var private_key = new byte[]
        {
            0x2, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1
        };
        var bad_signature = new byte[]
        {
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
            0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0
        };
        var kp = new KeyPair(private_key);
        var cg = new ContractGroup() { PubKey = kp.PublicKey, Signature = bad_signature};
        Console.Write(cg.ToJson());
        """
        bad_signature = b'\x00' * 64
        cg = contracts.ContractGroup(self.keypair.public_key, bad_signature)
        self.assertFalse(cg.is_valid(types.UInt160.zero()))

        # finally test is_valid() with a keypair we know the private key off
        contract_hash = b'\x01' * 20
        signature = cryptography.sign(contract_hash, self.keypair.private_key)

        cg2 = contracts.ContractGroup(self.keypair.public_key, signature)
        self.assertTrue(cg2.is_valid(types.UInt160(contract_hash)))

    def test_json(self):
        bad_signature = b'\x00' * 64
        cg = contracts.ContractGroup(self.keypair.public_key, bad_signature)
        self.assertEqual(self.expected_json, cg.to_json())

        # if the test_to_json() passes then we know our manually created ControlGroup (cg) is valid,
        # we can now use that to validate our `from_json` object
        cg2 = contracts.ContractGroup.from_json(self.expected_json)
        self.assertEqual(cg.public_key, cg2.public_key)
        self.assertEqual(cg.signature, cg2.signature)

    def test_eq(self):
        cg = contracts.ContractGroup.from_json(self.expected_json)
        cg2 = contracts.ContractGroup.from_json(self.expected_json)
        self.assertFalse(cg == object())
        self.assertTrue(cg == cg2)


class ContractPermissionTestCase(unittest.TestCase):
    def test_default(self):
        cp = contracts.ContractPermission.default_permissions()
        self.assertTrue(cp.contract.is_wildcard)
        self.assertTrue(cp.methods.is_wildcard)

    def test_is_allowed_based_on_hash(self):
        # We test the group permissions where all methods are allowed to be called
        # if contract_hash is valid.
        dummy_contract_hash = types.UInt160.zero()
        contract_state = contracts.ContractState(1,
                                               contracts.NEF(),
                                               contracts.ContractManifest(),
                                               0,
                                               dummy_contract_hash)

        # setup an allowed permission for a contract with UInt160.zero hash for all methods
        cpd = contracts.ContractPermissionDescriptor(contract_hash=dummy_contract_hash)
        cp = contracts.ContractPermission(contract=cpd, methods=contracts.WildcardContainer.create_wildcard())
        self.assertTrue(cp.is_allowed(contract_state, "dummy_method"))

        # now create a different contract hash and verify it does not give permission
        contract_state.hash = types.UInt160(b'\x01' * 20)
        self.assertFalse(cp.is_allowed(contract_state, "dummy_method"))

    def test_is_allowed_based_on_group(self):
        # We test the group permissions where all methods are allowed to be called
        # if the 'groups' member is valid.

        private_key = b'\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
        keypair = cryptography.KeyPair(private_key)

        dummy_contract_hash = types.UInt160.from_string("01" * 20)
        contract_state = contracts.ContractState(1,
                                               contracts.NEF(),
                                               contracts.ContractManifest(),
                                               0,
                                               dummy_contract_hash)

        signature = cryptography.sign(contract_state.hash.to_array(), keypair.private_key)
        contract_state.manifest.groups = [contracts.ContractGroup(keypair.public_key, signature)]

        cpd = contracts.ContractPermissionDescriptor(group=keypair.public_key)
        cp = contracts.ContractPermission(contract=cpd, methods=contracts.WildcardContainer.create_wildcard())
        self.assertTrue(cp.is_allowed(contract_state, "dummy_method"))

        # now modify the manifest to have a different `groups` attribute such that validation fails
        public_key = cryptography.ECPoint.deserialize_from_bytes(b'\x00')  # ECPoint.Infinity
        contract_state.manifest.groups = [contracts.ContractGroup(public_key, signature)]
        self.assertFalse(cp.is_allowed(contract_state, "dummy_method"))

    def test_is_allowed_invalid_method(self):
        # in the above tests we validated the 'group' and 'contract_hash' matching logic
        # now we validate 'method' matching
        dummy_contract_hash = types.UInt160.from_string("01" * 20)
        contract_state = contracts.ContractState(1,
                                               contracts.NEF(),
                                               contracts.ContractManifest(),
                                               0,
                                               dummy_contract_hash)

        # setup an allowed permission for a contract with UInt160.zero hash for 2 methods
        cpd = contracts.ContractPermissionDescriptor(contract_hash=dummy_contract_hash)
        cp = contracts.ContractPermission(contract=cpd, methods=contracts.WildcardContainer(data=['method1', 'method2']))
        self.assertTrue(cp.is_allowed(contract_state, "method1"))
        self.assertTrue(cp.is_allowed(contract_state, "method2"))
        self.assertFalse(cp.is_allowed(contract_state, "method3"))

    def test_to_json(self):
        # var cg = ContractPermission.DefaultPermission;
        # Console.WriteLine(cg.ToJson());
        cp = contracts.ContractPermission.default_permissions()
        expected = {"contract": "*", "methods": "*"}
        self.assertDictEqual(expected, cp.to_json())

    def test_from_json(self):
        # we use data from 'test_to_json'
        # we only test parsing basic data, as the detailed parsing is covered
        # by the tests for specific object classes
        json = {"contract": "*", "methods": "*"}
        from_json = contracts.ContractPermission.from_json(json)
        self.assertTrue(from_json.methods.is_wildcard)
        self.assertTrue(from_json.contract.is_wildcard)

    def test_eq(self):
        json = {"contract": "*", "methods": "*"}
        from_json = contracts.ContractPermission.from_json(json)
        from_json2 = contracts.ContractPermission.from_json(json)
        self.assertFalse(from_json == object())
        self.assertTrue(from_json == from_json2)


class WildcardContainerTestCase(unittest.TestCase):
    def test_dunders(self):
        wc = contracts.WildcardContainer(data=['method1', 'method2'])
        wc2 = contracts.WildcardContainer(data=['method1', 'method2'])
        self.assertIn('method1', wc)
        self.assertIn('method2', wc)
        self.assertNotIn('method3', wc)
        self.assertEqual('method2', wc[1])
        self.assertEqual(2, len(wc))
        self.assertNotEqual(wc, object())
        self.assertEqual(wc, wc2)

    def test_wildcard(self):
        wc = contracts.WildcardContainer.create_wildcard()
        self.assertTrue(wc.is_wildcard)

        wc = contracts.WildcardContainer(['method1'])
        self.assertFalse(wc.is_wildcard)

    def test_to_json(self):
        wc = contracts.WildcardContainer.create_wildcard()
        self.assertDictEqual({'wildcard': '*'}, wc.to_json())

        wc = contracts.WildcardContainer(data=['method1', 'method2'])
        self.assertDictEqual({'wildcard': ['method1', 'method2']}, wc.to_json())

    def test_from_json_default(self):
        wc = contracts.WildcardContainer.from_json({'wildcard': '*'})
        self.assertTrue(wc.is_wildcard)

        wc = contracts.WildcardContainer.from_json({'wildcard': ['method1', 'method2']})
        self.assertFalse(wc.is_wildcard)
        self.assertIn('method1', wc)
        self.assertIn('method2', wc)

        with self.assertRaises(ValueError) as context:
            contracts.WildcardContainer.from_json({})
        self.assertEqual("Invalid JSON - Cannot recreate wildcard from None", str(context.exception))

        with self.assertRaises(ValueError) as context:
            contracts.WildcardContainer.from_json({'wildcard': 'abc'})
        self.assertEqual("Invalid JSON - Cannot deduce WildcardContainer type from: abc", str(context.exception))

    def test_from_json_as_type(self):
        wc = contracts.WildcardContainer.from_json_as_type({'wildcard': '*'}, lambda: None)
        self.assertTrue(wc.is_wildcard)

        t1 = types.UInt160.zero()
        t2 = types.UInt160.from_string("11" * 20)
        t3 = types.UInt160.from_string("22" * 20)
        wc = contracts.WildcardContainer.from_json_as_type({'wildcard': [str(t1), str(t2)]},
                                                           lambda t: types.UInt160.from_string(t))
        self.assertFalse(wc.is_wildcard)
        self.assertIn(t1, wc)
        self.assertIn(t2, wc)
        self.assertNotIn(t3, wc)

        with self.assertRaises(ValueError) as context:
            contracts.WildcardContainer.from_json_as_type({}, lambda: None)
        self.assertEqual("Invalid JSON - Cannot recreate wildcard from None", str(context.exception))

        with self.assertRaises(ValueError) as context:
            contracts.WildcardContainer.from_json_as_type({'wildcard': 'abc'}, lambda: None)
        self.assertEqual("Invalid JSON - Cannot deduce WildcardContainer type from: abc", str(context.exception))


class ManifestTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var manifest = new ContractManifest()
        {
            Name = "test_contract",
            Groups = new ContractGroup[0],
            SupportedStandards = Array.Empty<string>(),
            Abi = new ContractAbi()
            {
                Events = new ContractEventDescriptor[0],
                Methods = new ContractMethodDescriptor[]
                {
                    new ContractMethodDescriptor {
                        Name = "main_entry",
                        Offset = 0,
                        Parameters = new ContractParameterDefinition[0],
                        ReturnType = ContractParameterType.Integer,
                        Safe = true
                    }
                }
            },
            Permissions = new[] { ContractPermission.DefaultPermission },
            Trusts = WildcardContainer<ContractPermissionDescriptor>.Create(),
            Extra = null
        };
        Console.WriteLine($"{manifest.ToJson()}");
        """
        cls.expected_json = {"name":"test_contract","groups":[],"features":{},"supportedstandards":[],"abi":{"methods":[{"name":"main_entry","parameters":[],"returntype":"Integer","offset":0,"safe":True}],"events":[]},"permissions":[{"contract":"*","methods":"*"}],"trusts":[],"extra":None}

    def test_create(self):
        cm = contracts.ContractManifest("test_contract")
        method1 = contracts.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=contracts.ContractParameterType.INTEGER,
            safe=True
        )
        cm.abi.methods = [method1]
        self.assertEqual(self.expected_json, cm.to_json())

    def test_serialize(self):
        # if test_create() passes, then we know `to_json()` is ok, which serialize internally uses
        cm = contracts.ContractManifest("test_contract")
        method1 = contracts.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=contracts.ContractParameterType.INTEGER,
            safe=True
        )
        cm.abi.methods = [method1]
        with serialization.BinaryReader(cm.to_array()) as br:
            data = br.read_var_string()
        self.assertDictEqual(self.expected_json, json.loads(data))

    def test_to_json_with_trusts_extra(self):
        # create a default manifest
        m = contracts.ContractManifest("test_contract")
        method1 = contracts.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=contracts.ContractParameterType.INTEGER,
            safe=True
        )
        m.abi.methods = [method1]

        t1 = contracts.ContractPermissionDescriptor(contract_hash=types.UInt160.from_string("01" * 20))
        t2 = contracts.ContractPermissionDescriptor(contract_hash=types.UInt160.from_string("02" * 20))
        m.trusts = contracts.WildcardContainer(data=[t1, t2])
        m.extra = False
        json_out = m.to_json()
        self.assertIn(t1.to_json()['contract'], json_out['trusts'])
        self.assertIn(t2.to_json()['contract'], json_out['trusts'])
        self.assertFalse(json_out['extra'])

    def test_from_json(self):
        expected_json = deepcopy(self.expected_json)
        cm = contracts.ContractManifest.from_json(expected_json)
        default = contracts.ContractManifest()
        self.assertEqual(default.groups, cm.groups)
        self.assertEqual(default.permissions, cm.permissions)
        self.assertEqual(default.trusts, cm.trusts)
        self.assertEqual(default.extra, cm.extra)

    def test_deserialization(self):
        # this assumes test_serialization() passes
        cm = contracts.ContractManifest("test_contract")
        method1 = contracts.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=contracts.ContractParameterType.INTEGER,
            safe=True
        )
        cm.abi.methods = [method1]
        cm_deserialized = contracts.ContractManifest.deserialize_from_bytes(cm.to_array())
        self.assertEqual(cm, cm_deserialized)

    def test_is_valid(self):
        # create a contract
        cm = contracts.ContractManifest("test_contract")
        method1 = contracts.ContractMethodDescriptor(
            name="main_entry",
            offset=0,
            parameters=[],
            return_type=contracts.ContractParameterType.INTEGER,
            safe=True
        )
        cm.abi.methods = [method1]

        # A contract hash is normally created from a combination of transaction sender (UInt160), NEF checksum and manifest.name
        # For this test we'll keep it simple and pretend that hash is all zeros
        dummy_contract_hash = types.UInt160.zero()

        # a "default" contract has no groups data, which is the same as always allow
        self.assertTrue(cm.is_valid(dummy_contract_hash))

        # now try to add a malicious group member (meaning; the member did not actually sign the ABI)
        private_key = b'\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
        keypair = cryptography.KeyPair(private_key)
        bad_signature = bytes(64)
        cm.groups = [contracts.ContractGroup(keypair.public_key, bad_signature)]
        # this time validation should fail
        self.assertFalse(cm.is_valid(dummy_contract_hash))

        # Finally test with a group member that did sign the ABI
        good_signature = cryptography.sign(dummy_contract_hash.to_array(), keypair.private_key)
        cm.groups = [contracts.ContractGroup(keypair.public_key, good_signature)]
        self.assertTrue(cm.is_valid(dummy_contract_hash))

    def test_eq(self):
        cm = contracts.ContractManifest.from_json(self.expected_json)
        cm2 = contracts.ContractManifest.from_json(self.expected_json)
        self.assertFalse(cm == object())
        self.assertTrue(cm == cm2)



from __future__ import annotations
import inspect
from typing import List, Callable, Dict, Any, Optional, get_type_hints
from neo3 import contracts, vm, storage, settings, _Singleton
from neo3.core import types, to_script_hash


class _NativeMethodMeta:
    def __init__(self, func: Callable):
        self.handler = func
        self.name: str = func.name  # type:ignore
        self.cpu_price: int = func.cpu_price  # type: ignore
        self.storage_price: int = func.storage_price  # type: ignore
        self.required_flags: contracts.CallFlags = func.flags  # type: ignore
        self.add_engine = False
        self.add_snapshot = False
        self.return_type = None

        parameter_types = []
        parameter_names = []
        for k, v in get_type_hints(func).items():
            if k == 'return':
                if v != type(None):
                    self.return_type = v
                continue
            if v == contracts.ApplicationEngine:
                self.add_engine = True
                continue
            elif v == storage.Snapshot:
                self.add_snapshot = True
                continue
            parameter_types.append(v)
            parameter_names.append(k)

        params = []
        for t, n in zip(parameter_types, parameter_names):
            params.append(contracts.ContractParameterDefinition(
                name=n,
                type=contracts.ContractParameterType.from_type(t)
            ))
        self.parameter_types = parameter_types
        self.descriptor = contracts.ContractMethodDescriptor(
            name=func.name,  # type: ignore
            offset=0,
            return_type=contracts.ContractParameterType.from_type(self.return_type),
            parameters=params,
            safe=(func.flags & ~contracts.CallFlags.READ_ONLY) == 0  # type: ignore
        )


class NativeContract(_Singleton):
    #: unique contract identifier
    _id: int = -99999

    #: A dictionary of all native contracts in the system
    _contracts: Dict[str, NativeContract] = {}
    #: A dictionary for accessing a native contract by its hash
    _contract_hashes: Dict[types.UInt160, NativeContract] = {}

    #: Allows for overriding the contract name in the ABI. Otherwise the name equals the class name.
    _service_name: Optional[str] = None

    #: The block index at which the native contract becomes active.
    active_block_index = 0

    def init(self):
        self._methods: Dict[int, _NativeMethodMeta] = {}  # offset, meta

        self._management = contracts.ManagementContract()
        self._neo = contracts.NeoToken()
        self._gas = contracts.GasToken()
        self._policy = contracts.PolicyContract()
        self._oracle = contracts.OracleContract()
        self._ledger = contracts.LedgerContract()
        self._role = contracts.DesignationContract()
        self._crypto = contracts.CryptoContract()
        self._stdlib = contracts.StdLibContract()

        # Find all methods that have been augmented by the @register decorator
        # and turn them into methods that can be called by VM scripts
        methods_meta = []
        for pair in inspect.getmembers(self, lambda m: hasattr(m, "native_call")):
            methods_meta.append(_NativeMethodMeta(pair[1]))

        methods_meta.sort(key=lambda x: (x.descriptor.name, len(x.descriptor.parameters)))

        sb = vm.ScriptBuilder()
        for meta in methods_meta:
            meta.descriptor.offset = len(sb)
            sb.emit_push(0)
            self._methods.update({len(sb): meta})
            sb.emit_syscall(1736177434)  # "System.Contract.CallNative"
            sb.emit(vm.OpCode.RET)

        self._script: bytes = sb.to_array()
        self.nef = contracts.NEF("neo-core-v3.0", self._script)

        sender = types.UInt160.zero()  # OpCode.PUSH1
        sb = vm.ScriptBuilder()
        sb.emit(vm.OpCode.ABORT)
        sb.emit_push(sender.to_array())
        sb.emit_push(0)
        sb.emit_push(self.service_name())
        self._hash: types.UInt160 = to_script_hash(sb.to_array())
        self._manifest: contracts.ContractManifest = contracts.ContractManifest()
        self._manifest.name = self.service_name()
        self._manifest.abi.methods = list(map(lambda m: m.descriptor, methods_meta))

        if self._id != NativeContract._id:
            self._contracts.update({self.service_name(): self})
            self._contract_hashes.update({self._hash: self})

        self.active_block_index = settings.native_contract_activation.get(self.service_name, 0)

    @classmethod
    def get_contract_by_name(cls, name: str) -> Optional[NativeContract]:
        """
        Get the contract instance by its service name
        Args:
            name: service name of the contract

        """
        contract = cls._contracts.get(name, None)
        return contract

    @classmethod
    def get_contract_by_id(cls, contract_id: int) -> Optional[NativeContract]:
        """ Get the native contract by its service id """
        for contract in cls._contracts.values():
            if contract_id == contract.id:
                return contract
        else:
            return None

    @classmethod
    def get_contract_by_hash(cls, contract_hash: types.UInt160) -> Optional[NativeContract]:
        """ Get the native contract by its contract hash """
        return cls._contract_hashes.get(contract_hash, None)

    @property
    def registered_contract_names(self) -> List[str]:
        """ The names of all deployed contracts. """
        return list(self._contracts.keys())

    @property
    def registered_contracts(self) -> List[NativeContract]:
        """ All deployed contracts. """
        return list(self._contracts.values())

    def service_name(self) -> str:
        """ The human readable name. """
        if self._service_name is None:
            return self.__class__.__name__
        else:
            return self._service_name

    @property
    def script(self) -> bytes:
        """ The contract byte code. """
        return self._script

    @property
    def hash(self) -> types.UInt160:
        """ Contract hash based of the contracts byte code + sender. """
        return self._hash

    @property
    def id(self) -> int:
        """ Unique identifier. """
        return self._id

    @property
    def manifest(self) -> contracts.ContractManifest:
        """ The associated contract manifest. """
        return self._manifest

    def invoke(self, engine: contracts.ApplicationEngine, version: int) -> None:
        """
        Calls a contract function

        Reads the required arguments from the engine's stack and converts them to the appropiate contract function types

        Args:
            engine: the engine executing the smart contract
            version: which version of the smart contract to load

        Raises:
             ValueError: if the request contract version is not
             ValueError: if the function to be called does not exist on the contract
             ValueError: if trying to call a function without having the correct CallFlags
        """
        if version != 0:
            raise ValueError(f"Native contract version {version} is not active")  # type: ignore

        context = engine.current_context
        flags = contracts.CallFlags(context.call_flags)
        method = self._methods.get(context.ip, None)
        if method is None:
            raise ValueError(f"Method at IP \"{context.ip}\" does not exist on contract {self.service_name()}")
        if method.required_flags not in flags:
            raise ValueError(f"Method requires call flag: {method.required_flags} received: {flags}")

        engine.add_gas(method.cpu_price
                       * contracts.PolicyContract().get_exec_fee_factor(engine.snapshot)
                       + method.storage_price
                       * contracts.PolicyContract().get_storage_price(engine.snapshot))

        params: List[Any] = []
        if method.add_engine:
            params.append(engine)

        if method.add_snapshot:
            params.append(engine.snapshot)

        for t in method.parameter_types:
            params.append(engine._stackitem_to_native(context.evaluation_stack.pop(), t))

        if len(params) > 0:
            return_value = method.handler(*params)
        else:
            return_value = method.handler()
        if method.return_type is not None:
            context.evaluation_stack.push(engine._native_to_stackitem(return_value, type(return_value)))

    @staticmethod
    def is_native(hash_: types.UInt160) -> bool:
        """
        Determine if the hash belong to a native contract.

        Args:
            hash_: a script hash

        Returns:
            True: if a contract is found matching the hash. False otherwise.
        """
        return hash_ in NativeContract._contract_hashes

    def on_persist(self, engine: contracts.ApplicationEngine) -> None:
        """
        Called by the Blockchain class when persisting a block.
        It will trigger the minting of GAS and updating of the NextValidators in storage

        Should not be called manually.
        """
        pass

    def post_persist(self, engine: contracts.ApplicationEngine):
        pass

    def create_key(self, prefix: bytes) -> storage.StorageKey:
        """
        Helper to create a storage key for the contract

        Args:
            prefix: the storage prefix to be used
        """
        return storage.StorageKey(self._id, prefix)

    def _initialize(self, engine: contracts.ApplicationEngine) -> None:
        """
        Called once when a native contract is deployed

        Args:
            engine: ApplicationEngine
        """

    def _check_committee(self, engine: contracts.ApplicationEngine) -> bool:
        addr = contracts.NeoToken().get_committee_address(engine.snapshot)
        return engine.checkwitness(addr)



"""The setup script."""
from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop
from contextlib import suppress
from pkg_resources import parse_version
import sys


class InstallCommand(install):
    user_options = install.user_options + [
        ('without-leveldb', None, 'Do not install leveldb requirements'),
    ]

    def initialize_options(self):
        super().initialize_options()
        # Initialize options
        self.without_leveldb = None

    def run(self):        # Use options
        if self.without_leveldb:
            print("install command called!")
            with suppress(ValueError):
                idx = list(map(lambda i: "plyvel" in i, self.distribution.install_requires)).index(True)
                self.distribution.install_requires.pop(idx)

        super().run()


class DevelopCommand(develop):
    user_options = develop.user_options + [
        ('without-leveldb', None, 'Do not install leveldb requirements'),
    ]

    def initialize_options(self):
        super().initialize_options()
        # Initialize options
        self.without_leveldb = None

    def run(self):
        # Use options
        if self.without_leveldb:
            with suppress(ValueError):
                idx = list(map(lambda i: "plyvel" in i, self.distribution.install_requires)).index(True)
                self.distribution.install_requires.pop(idx)

        super().run()

try:
    from pip._internal.req import parse_requirements
    from pip import __version__ as __pip_version
    pip_version = parse_version(__pip_version)
    if (pip_version >= parse_version("20")):
        from pip._internal.network.session import PipSession
    elif (pip_version >= parse_version("10")):
        from pip._internal.download import PipSession
except ImportError:  # pip version < 10.0
    from pip.req import parse_requirements
    from pip.download import PipSession

with open('README.rst') as readme_file:
    readme = readme_file.read()


leveldb_requirements = ["plyvel==1.3.0" if sys.platform in ["darwin", "linux"] else "plyvel-win32"]

# get the requirements from requirements.txt
install_reqs = parse_requirements('requirements.txt', session=PipSession())
reqs = []
for ir in install_reqs:
    if hasattr(ir, 'requirement'):
        reqs.append(str(ir.requirement))
    else:
        reqs.append(str(ir.req))
setup(
    name='neo-mamba',
    python_requires='>=3.8.*,<3.10',
    version='0.9.3',
    description="Python SDK for the NEO 3 blockchain",
    long_description=readme,
    long_description_content_type="text/x-rst",
    author="Erik van den Brink",
    author_email='erik@coz.io',
    maintainer="Erik van den Brink",
    maintainer_email='erik@coz.io',
    url='https://github.com/CityOfZion/neo-mamba',
    packages=find_packages(include=['neo3']),
    include_package_data=True,
    install_requires=reqs,
    extras_require={"leveldb": leveldb_requirements},
    license="MIT license",
    zip_safe=False,
    keywords='neo3, python, SDK',
    entry_points={
        'sphinx.html_themes': [
            'neo3 = docs.source._theme',
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
    ],
    cmdclass={
        'install': InstallCommand,
        'develop': DevelopCommand
    }
)



""" A module for filtering IPs via black and whitelists.

A global instance ``ipfilter`` can be imported directly from the module and is taken into account by default in the
NeoNode class when connections are established.

Filtering rules

* The whitelist has precedence over the blacklist settings.
* Host masks can be applied.
* When using host masks do not set host bits (leave them to 0) or an exception will occur.

.. _IPFilter examples:
Examples:
    The following are `configuration` examples for common scenario's.

    1. Accept only specific trusted IPs
    ::

        {
            'blacklist': [
                '0.0.0.0/0'
            ],
            'whitelist': [
                '10.10.10.10',
                '15.15.15.15'
            ]
        }

    2. Accept only a range of trusted IPs
    ::

        # accepts any IP in the range of 10.10.10.0 - 10.10.10.255
        {
            'blacklist': [
                '0.0.0.0/0'
            ],
            'whitelist': [
                '10.10.10.0/24',
            ]
        }

    3. Accept all except specific IPs
    ::

        # can be used for banning bad actors
        {
            'blacklist': [
                '12.12.12.12',
                '13.13.13.13'
            ],
            'whitelist': [
            ]
        }

"""
from ipaddress import IPv4Network
from contextlib import suppress
from typing import Dict, List
from copy import deepcopy


class IPFilter():

    default_config: dict = {'blacklist': [], 'whitelist': []}

    def __init__(self):
        self._config = deepcopy(self.default_config)

    def is_allowed(self, address: str) -> bool:
        """
        Test if a given address passes the configured restrictions.

        Args:
            address: an IPv4 address as defined in the :py:class:`standard library <python:ipaddress.IPv4Network>`.
        """
        ipv4_address = IPv4Network(address)

        is_allowed = True

        for ip in self._config['blacklist']:
            disallowed = IPv4Network(ip)
            if disallowed.overlaps(ipv4_address):
                is_allowed = False
                break
        else:
            return is_allowed

        # can override blacklist
        for ip in self._config['whitelist']:
            allowed = IPv4Network(ip)
            if allowed.overlaps(ipv4_address):
                is_allowed = True

        return is_allowed

    def blacklist_add(self, address: str) -> None:
        """
        Add an address that will not pass restriction checks.

        Args:
            address: an IPv4 address as defined in the :py:class:`standard library <python:ipaddress.IPv4Network>`.
        """
        self._config['blacklist'].append(address)

    def blacklist_remove(self, address: str) -> None:
        """
        Remove an address from the blacklist.

        Args:
            address: an IPv4 address as defined in the :py:class:`standard library <python:ipaddress.IPv4Network>`.
        """
        with suppress(ValueError):
            self._config['blacklist'].remove(address)

    def whitelist_add(self, address: str) -> None:
        """
        Add an address that will pass restriction checks.

        Args:
            address: an IPv4 address as defined in the :py:class:`standard library <python:ipaddress.IPv4Network>`.
        """
        self._config['whitelist'].append(address)

    def whitelist_remove(self, address: str) -> None:
        """
        Remove an address from the whitelist.

        Args:
            address: an IPv4 address as defined in the :py:class:`standard library <python:ipaddress.IPv4Network>`.
        """
        with suppress(ValueError):
            self._config['whitelist'].remove(address)

    def load_config(self, config: Dict[str, List[str]]) -> None:
        """
        Load filtering rules from a configuration object.

        Args:
            config: a _dictionary holding 2 keys, `blacklist` & `whitelist`, each  having a
                    :py:class:`list <python:list>` type value holding :py:class:`str <python:str>` type ``address`` es.
                    See :ref:`IPFilter examples`. For ``address`` format refer to the
                    :py:class:`standard library <python:ipaddress.IPv4Network>`.
        Raises:
            ValueError: if the required config keys are not found.
        """
        if 'whitelist' not in config:
            raise ValueError("whitelist key not found")
        if 'blacklist' not in config:
            raise ValueError("blacklist key not found")
        self._config = config

    def reset(self) -> None:
        """
        Clear the filter rules.
        """
        self._config = deepcopy(self.default_config)


ipfilter = IPFilter()



import unittest
from neo3 import settings, storage, vm, contracts, blockchain
from neo3.core import syscall_name_to_int, types, to_script_hash
from tests.contracts.interop.utils import test_engine, test_block, TestIVerifiable

# this triggers deployment of the native contracts
blockchain.Blockchain()


def contract_hash(sender: types.UInt160, checksum: int, name: str) -> types.UInt160:
    sb = vm.ScriptBuilder()
    sb.emit(vm.OpCode.ABORT)
    sb.emit_push(sender.to_array())
    sb.emit_push(checksum)
    sb.emit_push(name)
    return to_script_hash(sb.to_array())


def test_native_contract(contract_hash: types.UInt160, operation: str, args=None):
    engine = test_engine(has_snapshot=True)
    block = test_block(0)
    # or we won't pass the native deploy call
    engine.snapshot.persisting_block = block

    sb = vm.ScriptBuilder()
    sb.emit_dynamic_call(contract_hash, operation)

    script = vm.Script(sb.to_array())
    engine.load_script(script)

    # storing the current script in a contract otherwise "System.Contract.Call" will fail its checks
    nef = contracts.NEF(script=sb.to_array())
    manifest = contracts.ContractManifest("test_contract")
    next_id = contracts.ManagementContract().get_next_available_id(engine.snapshot)
    contract = contracts.ContractState(next_id + 1, nef, manifest, 0, to_script_hash(nef.script))
    engine.snapshot.contracts.put(contract)

    return engine


class TestPolicyContract(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        settings.network.standby_committee = ['02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765']
        settings.network.validators_count = 1

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()

    def test_basics(self):
        policy = contracts.PolicyContract()
        self.assertEqual(-7, policy.id)
        self.assertEqual("PolicyContract", contracts.PolicyContract().service_name())

    def test_policy_default_get_fee_per_byte(self):
        engine = test_native_contract(contracts.PolicyContract().hash, "getFeePerByte")
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertEqual(vm.IntegerStackItem(1000), item)

    def test_policy_block_account_and_is_blocked(self):
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        # set or we won't pass the native deploy call
        engine.snapshot.persisting_block = block

        sb = vm.ScriptBuilder()

        # set or we won't pass the check_comittee() in the policy contract function implementations
        engine.script_container = TestIVerifiable()
        validator = settings.standby_committee[0]
        script_hash = to_script_hash(contracts.Contract.create_multisig_redeemscript(1, [validator]))
        engine.script_container.script_hashes = [script_hash]

        # first we setup the stack for calling `blockAccount`
        # push data to create a vm.Array holding 20 bytes for the UInt160 Account parameter of the _block_account function.
        sb.emit_push(b'\x11' * 20)
        sb.emit(vm.OpCode.PUSH1)
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(15)  # call flags
        sb.emit_push("blockAccount")
        sb.emit_push(contracts.PolicyContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))

        # next we call `isBlocked`
        sb.emit_push(b'\x11' * 20)
        sb.emit(vm.OpCode.PUSH1)
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(15)  # call flags
        sb.emit_push("isBlocked")
        sb.emit_push(contracts.PolicyContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))

        script = vm.Script(sb.to_array())
        engine.load_script(script)

        # storing the current script in a contract otherwise "System.Contract.Call" will fail its checks
        nef = contracts.NEF(script=sb.to_array())
        manifest = contracts.ContractManifest("test_contract")
        sender = engine.script_container.script_hashes[0]
        contract = contracts.ContractState(1, nef, manifest, 0, contract_hash(sender, nef.checksum, manifest.name))
        engine.snapshot.contracts.put(contract)

        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(2, len(engine.result_stack))
        get_is_blocked_result = engine.result_stack.pop()
        set_blocked_account_result = engine.result_stack.pop()
        self.assertTrue(set_blocked_account_result.to_boolean())
        self.assertTrue(get_is_blocked_result.to_boolean())

    def test_policy_unblock_account(self):
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        engine.snapshot.persisting_block = block

        # we must add a script_container with valid signature to pass the check_comittee() validation check
        # in the function itself
        engine.script_container = TestIVerifiable()
        validator = settings.standby_committee[0]
        script_hash = to_script_hash(contracts.Contract.create_multisig_redeemscript(1, [validator]))
        engine.script_container.script_hashes = [script_hash]

        policy = contracts.PolicyContract()
        account_not_found = types.UInt160(data=b'\x11' * 20)
        account = types.UInt160.zero()
        self.assertTrue(policy._block_account(engine, account))
        self.assertFalse(policy._unblock_account(engine, account_not_found))
        self.assertTrue(policy._unblock_account(engine, account))
        storage_key = policy.key_blocked_account + account
        storage_item = engine.snapshot.storages.try_get(storage_key)
        self.assertIsNone(storage_item)

    def test_policy_setters_fail_without_signatures(self):
        # cover set functions where check_committee fails
        policy = contracts.PolicyContract()
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        engine.snapshot.persisting_block = block
        engine.script_container = TestIVerifiable()

        with self.assertRaises(ValueError) as context:
            policy._set_fee_per_byte(engine, 0)
        self.assertEqual("Check committee failed", str(context.exception))

        self.assertFalse(policy._block_account(engine, None))
        self.assertFalse(policy._unblock_account(engine, None))



from __future__ import annotations
import struct
from enum import IntEnum
from typing import List
from . import NativeContract, register
from neo3 import storage, contracts, cryptography, vm
from neo3.core import serialization, msgrouter


class DesignateRole(IntEnum):
    STATE_VALIDATOR = 4
    ORACLE = 8
    NEO_FS_ALPHABET_NODE = 16


class DesignationContract(NativeContract):
    _id = -8
    _service_name = "RoleManagement"

    def init(self):
        super(DesignationContract, self).init()
        self.manifest.abi.events = [
            contracts.ContractEventDescriptor(
                "Designation",
                parameters=[
                    contracts.ContractParameterDefinition("Role", contracts.ContractParameterType.INTEGER),
                    contracts.ContractParameterDefinition("BlockIndex", contracts.ContractParameterType.INTEGER),
                ]
            ),
            contracts.ContractEventDescriptor(
                "OracleResponse",
                parameters=[
                    contracts.ContractParameterDefinition("Id", contracts.ContractParameterType.INTEGER),
                    contracts.ContractParameterDefinition("OriginalTx", contracts.ContractParameterType.HASH160)
                ]
            )
        ]

    @register("getDesignatedByRole", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_designated_by_role(self,
                               snapshot: storage.Snapshot,
                               role: DesignateRole,
                               index: int) -> List[cryptography.ECPoint]:
        if snapshot.best_block_height + 1 < index:
            raise ValueError("[DesignateContract] Designate list index out of range")

        key = self.create_key(role.to_bytes(1, 'little') + self._to_uint32(index)).to_array()
        boundary = self.create_key(role.to_bytes(1, 'little')).to_array()
        for _, storage_item in snapshot.storages.find_range(key, boundary, "reverse"):
            with serialization.BinaryReader(storage_item.value) as reader:
                return reader.read_serializable_list(cryptography.ECPoint)
        else:
            return []

    @register("designateAsRole", contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_NOTIFY, cpu_price=1 << 15)
    def designate_as_role(self,
                          engine: contracts.ApplicationEngine,
                          role: DesignateRole,
                          nodes: List[cryptography.ECPoint]) -> None:
        if len(nodes) == 0:
            raise ValueError("[DesignateContract] Cannot designate empty nodes list")

        if len(nodes) > 32:
            raise ValueError("[DesignateContract] Cannot designate a nodes list larger than 32")

        if not self._check_committee(engine):
            raise ValueError("[DesignateContract] check committee failed")

        if engine.snapshot.persisting_block is None:
            raise ValueError

        nodes.sort()
        index = engine.snapshot.persisting_block.index + 1
        storage_key = self.create_key(role.to_bytes(1, 'little') + self._to_uint32(index))
        with serialization.BinaryWriter() as writer:
            writer.write_serializable_list(nodes)
            storage_item = storage.StorageItem(writer.to_array())
        engine.snapshot.storages.update(storage_key, storage_item)

        state = vm.ArrayStackItem(engine.reference_counter)
        state.append(vm.IntegerStackItem(role.value))
        state.append(vm.IntegerStackItem(engine.snapshot.persisting_block.index))
        msgrouter.interop_notify(self.hash, "Designation", state)

    def _to_uint32(self, value: int) -> bytes:
        return struct.pack(">I", value)



from __future__ import annotations
import json
from . import NativeContract, register
from typing import Optional
from neo3 import storage, contracts, vm
from neo3.core import to_script_hash, types, msgrouter
from neo3.network import payloads


class ManagementContract(NativeContract):
    _id = -1
    _service_name = "ContractManagement"

    key_min_deploy_fee = storage.StorageKey(_id, b'\x14')
    key_next_id = storage.StorageKey(_id, b'\x0f')

    def init(self):
        super(ManagementContract, self).init()

        self.manifest.abi.events = [
            contracts.ContractEventDescriptor(
                "Deploy",
                parameters=[
                    contracts.ContractParameterDefinition("Hash", contracts.ContractParameterType.HASH160)
                ]
            ),
            contracts.ContractEventDescriptor(
                "Update",
                parameters=[
                    contracts.ContractParameterDefinition("Hash", contracts.ContractParameterType.HASH160)
                ]
            ),
            contracts.ContractEventDescriptor(
                "Destroy",
                parameters=[
                    contracts.ContractParameterDefinition("Hash", contracts.ContractParameterType.HASH160)
                ]
            ),
        ]

    def _initialize(self, engine: contracts.ApplicationEngine) -> None:
        engine.snapshot.storages.put(
            self.key_min_deploy_fee,
            storage.StorageItem(vm.BigInteger(10_00000000).to_array())
        )
        engine.snapshot.storages.put(self.key_next_id, storage.StorageItem(vm.BigInteger(1).to_array()))

    @register("getContract", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_contract(self, snapshot: storage.Snapshot, hash_: types.UInt160) -> Optional[contracts.ContractState]:
        return snapshot.contracts.try_get(hash_, read_only=True)

    @register("deploy", contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_NOTIFY)
    def contract_create(self,
                        engine: contracts.ApplicationEngine,
                        nef_file: bytes,
                        manifest: bytes) -> contracts.ContractState:
        return self.contract_create_with_data(engine, nef_file, manifest, vm.NullStackItem())

    @register("deploy", contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_NOTIFY)
    def contract_create_with_data(self,
                                  engine: contracts.ApplicationEngine,
                                  nef_file: bytes,
                                  manifest: bytes,
                                  data: vm.StackItem) -> contracts.ContractState:
        if not isinstance(engine.script_container, payloads.Transaction):
            raise ValueError("Cannot create contract without a Transaction script container")

        nef_len = len(nef_file)
        manifest_len = len(manifest)
        if (nef_len == 0
                or nef_len > engine.MAX_CONTRACT_LENGTH
                or manifest_len == 0
                or manifest_len > contracts.ContractManifest.MAX_LENGTH):
            raise ValueError("Invalid NEF or manifest length")

        engine.add_gas(
            max(engine.STORAGE_PRICE * (nef_len + manifest_len), self.get_minimum_deployment_fee(engine.snapshot))
        )

        nef = contracts.NEF.deserialize_from_bytes(nef_file)
        parsed_manifest = contracts.ContractManifest.from_json(json.loads(manifest.decode()))

        self.validate(nef.script, parsed_manifest.abi)

        sb = vm.ScriptBuilder()
        sb.emit(vm.OpCode.ABORT)
        sb.emit_push(engine.script_container.sender.to_array())
        sb.emit_push(nef.checksum)
        sb.emit_push(parsed_manifest.name)
        hash_ = to_script_hash(sb.to_array())

        existing_contract = engine.snapshot.contracts.try_get(hash_)
        if existing_contract is not None:
            raise ValueError("Contract already exists")

        contract = contracts.ContractState(self.get_next_available_id(engine.snapshot), nef, parsed_manifest, 0, hash_)
        if not contract.manifest.is_valid(hash_):
            raise ValueError("Error: invalid manifest")
        engine.snapshot.contracts.put(contract)

        method_descriptor = contract.manifest.abi.get_method("_deploy", 2)
        if method_descriptor is not None:
            engine.call_from_native(self.hash, hash_, method_descriptor.name, [data, vm.BooleanStackItem(False)])

        msgrouter.interop_notify(self.hash,
                                 "Deploy",
                                 vm.ArrayStackItem(engine.reference_counter,
                                                   vm.ByteStringStackItem(contract.hash.to_array())
                                                   )
                                 )
        return contract

    @register("update", contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_NOTIFY)
    def contract_update(self,
                        engine: contracts.ApplicationEngine,
                        nef_file: Optional[bytes],
                        manifest: Optional[bytes]) -> None:
        self.contract_update_with_data(engine, nef_file, manifest, vm.NullStackItem())

    @register("update", contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_NOTIFY)
    def contract_update_with_data(self,
                                  engine: contracts.ApplicationEngine,
                                  nef_file: Optional[bytes],
                                  manifest: Optional[bytes],
                                  data: vm.StackItem) -> None:
        if nef_file is None and manifest is None:
            raise ValueError("Cannot upgrade contract without a new NEF or a new manifest")

        nef_len = 0 if nef_file is None else len(nef_file)
        manifest_len = 0 if manifest is None else len(manifest)
        engine.add_gas(engine.STORAGE_PRICE * (nef_len + manifest_len))

        contract = engine.snapshot.contracts.try_get(engine.calling_scripthash, read_only=False)
        if contract is None:
            raise ValueError("Can't find contract to update")

        if nef_file is not None:
            if nef_len == 0:
                raise ValueError(f"Invalid NEF length: {nef_len}")

            # update contract
            contract.nef = contracts.NEF.deserialize_from_bytes(nef_file)

        if manifest is not None:
            if manifest_len == 0 or manifest_len > contracts.ContractManifest.MAX_LENGTH:
                raise ValueError(f"Invalid manifest length: {manifest_len}")

            manifest_new = contracts.ContractManifest.from_json(json.loads(manifest.decode()))
            if manifest_new.name != contract.manifest.name:
                raise ValueError("Error: cannot change contract name")
            if not contract.manifest.is_valid(contract.hash):
                raise ValueError("Error: manifest does not match with script")
            contract.manifest = manifest_new

        self.validate(contract.nef.script, contract.manifest.abi)

        contract.update_counter += 1

        method_descriptor = contract.manifest.abi.get_method("_deploy", 2)
        if method_descriptor is not None:
            engine.call_from_native(self.hash,
                                    contract.hash,
                                    method_descriptor.name,
                                    [data, vm.BooleanStackItem(True)])

        msgrouter.interop_notify(self.hash,
                                 "Update",
                                 vm.ArrayStackItem(engine.reference_counter,
                                                   vm.ByteStringStackItem(contract.hash.to_array())
                                                   )
                                 )

    @register("destroy", contracts.CallFlags.STATES | contracts.CallFlags.ALLOW_NOTIFY, cpu_price=1 << 15)
    def contract_destroy(self, engine: contracts.ApplicationEngine) -> None:
        hash_ = engine.calling_scripthash
        contract = engine.snapshot.contracts.try_get(hash_)

        if contract is None:
            return

        engine.snapshot.contracts.delete(hash_)

        for key, _ in engine.snapshot.storages.find(contract.id.to_bytes(4, 'little', signed=True)):
            engine.snapshot.storages.delete(key)

        msgrouter.interop_notify(self.hash,
                                 "Destroy",
                                 vm.ArrayStackItem(engine.reference_counter,
                                                   vm.ByteStringStackItem(contract.hash.to_array())
                                                   )
                                 )

    @register("getMinimumDeploymentFee", contracts.CallFlags.READ_STATES, cpu_price=1 << 15)
    def get_minimum_deployment_fee(self, snapshot: storage.Snapshot) -> int:
        return int.from_bytes(snapshot.storages.get(self.key_min_deploy_fee, read_only=True).value, 'little')

    @register("setMinimumDeploymentFee", contracts.CallFlags.STATES, cpu_price=1 << 15)
    def _set_minimum_deployment_fee(self, engine: contracts.ApplicationEngine, value: int) -> None:
        if value < 0:
            raise ValueError("Can't set deployment fee to a negative value")
        if not self._check_committee(engine):
            raise ValueError
        engine.snapshot.storages.update(self.key_min_deploy_fee, storage.StorageItem(vm.BigInteger(value).to_array()))

    def get_next_available_id(self, snapshot: storage.Snapshot) -> int:
        si = snapshot.storages.get(self.key_next_id, read_only=False)
        value = vm.BigInteger(si.value)
        si.value = (value + 1).to_array()
        return int(value)

    def on_persist(self, engine: contracts.ApplicationEngine) -> None:
        # NEO implicitely expects a certain order of contract initialization
        # Native contracts have negative values for `id`, so we reverse the results
        sorted_contracts = sorted(self.registered_contracts, key=lambda contract: contract.id, reverse=True)
        for contract in sorted_contracts:
            if contract.active_block_index != engine.snapshot.persisting_block.index:
                continue
            engine.snapshot.contracts.put(
                contracts.ContractState(contract.id, contract.nef, contract.manifest, 0, contract.hash)
            )
            contract._initialize(engine)

    def validate(self, script: bytes, abi: contracts.ContractABI):
        s = vm.Script(script, True)
        for method in abi.methods:
            s.get_instruction(method.offset)
        events = []
        for event in abi.events:
            if event.name in events:
                raise ValueError("Duplicate event in ABI")
            else:
                events.append(event)



from __future__ import annotations
from contextlib import suppress
from copy import deepcopy
from typing import List, Iterator, Tuple, Optional, TYPE_CHECKING
from neo3 import storage, storage_logger as logger
from neo3.core import types, serialization
from neo3.network import payloads
from neo3 import contracts


level_db_supported = False
with suppress(ModuleNotFoundError, ImportError):
    import plyvel  # type: ignore
    level_db_supported = True


class DBPrefixes:
    BLOCKS = b'\x01'
    BLOCKS_HEIGHT_MAP = b'\x02'
    BLOCKS_BEST_HEIGHT = b'\x03'
    CONTRACTS = b'\x04'
    STORAGES = b'\x06'
    TRANSACTIONS = b'\x07'


class LevelDB(storage.IDBImplementation):
    def __init__(self, options: dict):
        if not level_db_supported:
            raise ModuleNotFoundError("plyvel module not found - try 'pip install plyvel'. "
                                      "Also make sure to have leveldb installed.")
        try:
            self._path = options['path']
            self._real_db = plyvel.DB(self._path, create_if_missing=True, max_open_files=100,
                                      lru_cache_size=10 * 1024 * 1024)
            self._tx_iterator = None
            self._block_iterator = None
            self._contract_iterator = None
            logger.info(f"Created DB at {self._path}")
        except Exception as e:
            raise Exception(f"leveldb exception [ {e} ]")

    def get_snapshotview(self) -> LevelDBSnapshot:
        return LevelDBSnapshot(self)

    def _internal_bestblockheight_get(self):
        height_bytes = self._real_db.get(DBPrefixes.BLOCKS_BEST_HEIGHT)
        if height_bytes is None:
            raise KeyError
        return int.from_bytes(height_bytes, 'little')

    def _internal_bestblockheight_put(self, height: int, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        db.put(DBPrefixes.BLOCKS_BEST_HEIGHT, height.to_bytes(4, 'little'))

    def _internal_bestblockheight_update(self, height: int, batch=None):
        self._internal_bestblockheight_put(height, batch)

    def _internal_block_put(self, block: payloads.Block, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        block_height_bytes = block.index.to_bytes(4, 'little')
        db.put(DBPrefixes.BLOCKS + block.hash().to_array(), block.to_array())
        db.put(DBPrefixes.BLOCKS_HEIGHT_MAP + block_height_bytes, block.hash().to_array())

        # this function is only called when putting blocks to the backend via the raw view, or when committing
        # a snapshot view. Either way it is ok to persist the height
        stored_value = -1
        with suppress(KeyError):
            stored_value = self._internal_bestblockheight_get()

        if block.index > stored_value:
            db.put(DBPrefixes.BLOCKS_BEST_HEIGHT, block_height_bytes)

    def _internal_block_update(self, block: payloads.Block, batch=None):
        self._internal_block_put(block, batch)

    def _internal_block_delete(self, hash: types.UInt256, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        block_hash_bytes = hash.to_array()
        block_bytes = self._real_db.get(DBPrefixes.BLOCKS + block_hash_bytes)
        if block_bytes is not None:
            # Instead of full block deserialization, which includes merkletree calculation and such, we manually extract
            # the 4 bytes block.index from the stream.
            start_idx = 4 + 32 + 32 + 8
            block_height_bytes = block_bytes[start_idx:start_idx + 4]
            db.delete(DBPrefixes.BLOCKS + block_hash_bytes)
            db.delete(DBPrefixes.BLOCKS_HEIGHT_MAP + block_height_bytes)

    def _internal_block_get(self, hash: types.UInt256):
        block_bytes = self._real_db.get(DBPrefixes.BLOCKS + hash.to_array())
        if block_bytes is None:
            raise KeyError

        return payloads.Block.deserialize_from_bytes(block_bytes)

    def _internal_block_get_by_height(self, height: int):
        block_hash_bytes = self._real_db.get(DBPrefixes.BLOCKS_HEIGHT_MAP + height.to_bytes(4, 'little'))
        if block_hash_bytes is None:
            raise KeyError

        block_bytes = self._real_db.get(DBPrefixes.BLOCKS + block_hash_bytes)
        if block_bytes is None:
            # should not be reachable unless _internal_block_put/delete() are messed up.
            raise KeyError  # pragma: no cover

        return payloads.Block.deserialize_from_bytes(block_bytes)

    def _internal_block_all(self):
        res = []

        with self._real_db.iterator(prefix=DBPrefixes.BLOCKS, include_key=False, include_value=True) as it:
            for value in it:
                v = payloads.Block.deserialize_from_bytes(value)
                res.append(v)

            # yielding outside of iterator to make sure the LevelDB iterator is closed and not leaking resources
            for block in res:
                yield deepcopy(block)

    def _internal_contract_put(self, contract: contracts.ContractState, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        db.put(DBPrefixes.CONTRACTS + contract.hash.to_array(), contract.to_array())

    def _internal_contract_update(self, contract: contracts.ContractState, batch=None):
        self._internal_contract_put(contract, batch)

    def _internal_contract_delete(self, hash: types.UInt160, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        db.delete(DBPrefixes.CONTRACTS + hash.to_array())

    def _internal_contract_get(self, hash: types.UInt160):
        contract_bytes = self._real_db.get(DBPrefixes.CONTRACTS + hash.to_array())
        if contract_bytes is None:
            raise KeyError

        return contracts.ContractState.deserialize_from_bytes(contract_bytes)

    def _internal_contract_all(self):
        res = []
        with self._real_db.iterator(prefix=DBPrefixes.CONTRACTS, include_key=False, include_value=True) as it:
            for value in it:
                # strip off prefix
                v = contracts.ContractState.deserialize_from_bytes(value)
                res.append(v)

        # yielding outside of iterator to make sure the LevelDB iterator is closed and not leaking resources
        for contract in res:
            yield deepcopy(contract)

    def _internal_storage_put(self, key: storage.StorageKey, value: storage.StorageItem, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        db.put(DBPrefixes.STORAGES + key.to_array(), value.to_array())

    def _internal_storage_update(self, key: storage.StorageKey, value: storage.StorageItem, batch=None):
        self._internal_storage_put(key, value)

    def _internal_storage_delete(self, key: storage.StorageKey, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        db.delete(DBPrefixes.STORAGES + key.to_array())

    def _internal_storage_get(self, key: storage.StorageKey):
        storage_bytes = self._real_db.get(DBPrefixes.STORAGES + key.to_array())
        if storage_bytes is None:
            raise KeyError

        return storage.StorageItem.deserialize_from_bytes(storage_bytes)

    def _internal_storage_all(self, contract_id: Optional[int] = None):
        prefix = DBPrefixes.STORAGES
        if contract_id is not None:
            prefix = DBPrefixes.STORAGES + contract_id.to_bytes(4, 'little', signed=True)

        res = {}
        with self._real_db.iterator(prefix=prefix, include_key=True, include_value=True) as it:
            for key, value in it:
                # strip off prefix
                k = storage.StorageKey.deserialize_from_bytes(key[1:])
                v = storage.StorageItem.deserialize_from_bytes(value)
                res[k] = v

        # yielding outside of iterator to make sure the LevelDB iterator is closed and not leaking resources
        for k, v in res.items():
            yield deepcopy(k), deepcopy(v)

    def _internal_storage_find(self, key_prefix: bytes):
        prefix = DBPrefixes.STORAGES + key_prefix

        res = {}
        with self._real_db.iterator(prefix=prefix, include_key=True, include_value=True) as it:
            for key, value in it:
                # strip off STORAGES table prefix
                k = storage.StorageKey.deserialize_from_bytes(key[1:])
                v = storage.StorageItem.deserialize_from_bytes(value)
                res[k] = v

        # yielding outside of iterator to make sure the LevelDB iterator is closed and not leaking resources
        for k, v in res.items():
            yield k, v

    def _internal_storage_seek(self,
                               key_prefix: bytes,
                               seek_direction="forward") -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:

        target = DBPrefixes.STORAGES + key_prefix

        with self._real_db.raw_iterator() as it:
            if seek_direction == "forward":
                it.seek(target)
                while it.valid():
                    key = it.key()
                    if len(key) < 1 or key[0] != DBPrefixes.STORAGES[0]:
                        break

                    k = storage.StorageKey.deserialize_from_bytes(key[1:])
                    v = storage.StorageItem.deserialize_from_bytes(it.value())
                    yield (k, v)
                    it.next()
            else:
                it.seek(target)
                if not it.valid():
                    it.seek_to_last()
                elif it.key() > target:
                    it.prev()

                while it.valid():
                    key = it.key()
                    if len(key) < 1 or key[0] != DBPrefixes.STORAGES[0]:
                        break

                    k = storage.StorageKey.deserialize_from_bytes(key[1:])
                    v = storage.StorageItem.deserialize_from_bytes(it.value())
                    yield (k, v)
                    it.prev()

    def _internal_transaction_put(self, transaction: payloads.Transaction, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db

        with serialization.BinaryWriter() as bw:
            transaction.serialize_special(bw)
            serialized_tx = bw.to_array()

        db.put(DBPrefixes.TRANSACTIONS + transaction.hash().to_array(), serialized_tx)

    def _internal_transaction_update(self, transaction, batch=None):
        self._internal_transaction_put(transaction, batch)

    def _internal_transaction_delete(self, hash: types.UInt256, batch=None):
        if batch:
            db = batch
        else:
            db = self._real_db
        db.delete(DBPrefixes.TRANSACTIONS + hash.to_array())

    def _internal_transaction_get(self, hash: types.UInt256):
        tx_bytes = self._real_db.get(DBPrefixes.TRANSACTIONS + hash.to_array())
        if tx_bytes is None:
            # this is a must if not found!
            raise KeyError

        with serialization.BinaryReader(tx_bytes) as br:
            tx = payloads.Transaction._serializable_init()
            tx.deserialize_special(br)
            return tx

    def _internal_transaction_all(self):
        res = []
        with self._real_db.iterator(prefix=DBPrefixes.TRANSACTIONS, include_key=False, include_value=True) as it:
            for value in it:
                # strip off prefix
                with serialization.BinaryReader(value) as br:
                    v = payloads.Transaction._serializable_init()
                    v.deserialize_special(br)
                    res.append(v)

        # yielding outside of iterator to make sure the LevelDB iterator is closed and not leaking resources
        for tx in res:
            yield deepcopy(tx)

    def close(self):
        self._real_db.close()


class LevelDBSnapshot(storage.Snapshot):
    def __init__(self, db: LevelDB):
        super(LevelDBSnapshot, self).__init__()
        self._db = db
        self._snapshot = db._real_db.snapshot()
        self._batch = db._real_db.write_batch()

        self._block_cache = LevelDBCachedBlockAccess(db, self._batch)
        self._contract_cache = LevelDBCachedContractAccess(db, self._batch)
        self._storage_cache = LevelDBCachedStorageAccess(db, self._batch)
        self._tx_cache = LevelDBCachedTXAccess(db, self._batch)
        self._block_height_cache = LevelDBBestBlockHeightAttribute(db, self._batch)

    def commit(self) -> None:
        super(LevelDBSnapshot, self).commit()
        self._batch.write()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._snapshot.close()


class LevelDBBestBlockHeightAttribute(storage.AttributeCache):
    def __init__(self, db, batch):
        super(LevelDBBestBlockHeightAttribute, self).__init__()
        self._db = db
        self._batch = batch

    def _get_internal(self):
        return self._db._internal_bestblockheight_get()

    def _update_internal(self, value):
        self._db._internal_bestblockheight_update(value, self._batch)


class LevelDBCachedBlockAccess(storage.CachedBlockAccess):
    def __init__(self, db, batch):
        super(LevelDBCachedBlockAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[types.UInt256] = []
        for trackable in self._dictionary.values():  # trackable.item: payloads.Block
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_block_put(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_block_update(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_block_delete(trackable.item.hash(), self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneBlockCache(self._db, self)


class LevelDBCachedContractAccess(storage.CachedContractAccess):
    def __init__(self, db, batch):
        super(LevelDBCachedContractAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[types.UInt160] = []
        for trackable in self.get_changeset():  # trackable.item: contracts.ContractState
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_contract_put(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_contract_update(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_contract_delete(trackable.item.hash, self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneContractCache(self._db, self)


class LevelDBCachedStorageAccess(storage.CachedStorageAccess):
    def __init__(self, db, batch):
        super(LevelDBCachedStorageAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[storage.StorageKey] = []
        for trackable in self.get_changeset():
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_storage_put(trackable.key, trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_storage_update(trackable.key, trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_storage_delete(trackable.key, self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneStorageCache(self._db, self)


class LevelDBCachedTXAccess(storage.CachedTXAccess):
    def __init__(self, db, batch):
        super(LevelDBCachedTXAccess, self).__init__(db)
        self._batch = batch

    def commit(self) -> None:
        keys_to_delete: List[types.UInt256] = []
        for trackable in self._dictionary.values():  # trackable.item: payloads.Transaction
            if trackable.state == storage.TrackState.ADDED:
                self._db._internal_transaction_put(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.CHANGED:
                self._db._internal_transaction_update(trackable.item, self._batch)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == storage.TrackState.DELETED:
                self._db._internal_transaction_delete(trackable.item.hash(), self._batch)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()

    def create_snapshot(self):
        return storage.CloneTXCache(self._db, self)



from .flightinfo import FlightInfo
from .nodeweight import NodeWeight
from .requestinfo import RequestInfo
from .nodemanager import NodeManager
from .syncmanager import SyncManager


__all__ = ['FlightInfo', 'NodeWeight', 'RequestInfo', 'NodeManager', 'SyncManager']



from __future__ import annotations
import datetime
from typing import List
from neo3.core import Size as s, serialization, utils
from neo3.network import capabilities
from neo3 import settings


class VersionPayload(serialization.ISerializable):
    """
    A payload carrying node handshake data.
    """
    MAX_CAPABILITIES = 32

    def __init__(self, nonce: int, user_agent: str, capabilities: List[capabilities.NodeCapability]):
        """
        Create payload.

        Args:
            nonce: unique number which identifies the node instance.
            user_agent: node user agent description. e.g. "NEO3-PYTHON-V001". Max 1024 bytes.
            capabilities: a list of services a node offers.
        """

        #: A network id. Differs for NEO's Mainnet (use 5195086) and Testnet (use 1951352142).
        #:
        #: Reference nodes will disconnect if the value doesn't match with their local settings.
        self.magic = settings.network.magic
        #: Protocol version of the node
        self.version = 0
        #: The UTC time when the node connected
        self.timestamp = int(datetime.datetime.utcnow().timestamp())
        #: A unique identifier for the node.
        self.nonce = nonce
        #: A node client description i.e. "NEO-MAMBA-V001"
        self.user_agent = user_agent
        #: A list of services a node offers. See :ref:`capabilities <library-network-capabilities>`
        self.capabilities = capabilities

    def __len__(self):
        """ Get the total size in bytes of the object. """
        return s.uint32 + s.uint32 + s.uint32 + s.uint32 + utils.get_var_size(self.user_agent) + \
            utils.get_var_size(self.capabilities)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.magic)
        writer.write_uint32(self.version)
        writer.write_uint32(self.timestamp)
        writer.write_uint32(self.nonce)
        writer.write_var_string(self.user_agent)
        writer.write_serializable_list(self.capabilities)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.magic = reader.read_uint32()
        self.version = reader.read_uint32()
        self.timestamp = reader.read_uint32()
        self.nonce = reader.read_uint32()
        self.user_agent = reader.read_var_string(max=1024)

        capabilities_cnt = reader.read_var_int(self.MAX_CAPABILITIES)
        capabilities_list = []
        for _ in range(capabilities_cnt):
            capabilities_list.append(capabilities.NodeCapability.deserialize_from(reader))
        self.capabilities = capabilities_list

    @classmethod
    def _serializable_init(cls):
        return cls(0, "", [])



from __future__ import annotations
import abc
from neo3.core import types
from neo3 import storage
from typing import Tuple, Optional, Iterator, TYPE_CHECKING

if TYPE_CHECKING:
    from neo3.network import payloads
    from neo3 import contracts


class IDBImplementation(abc.ABC):
    def get_rawview(self):
        """ Get a view of the database without caching. """
        return RawView(self)

    @abc.abstractmethod
    def get_snapshotview(self) -> storage.Snapshot:
        """ Get a snapshot view of the database with caching. """

    def close(self):
        """ clean up any resources."""

    @abc.abstractmethod
    def _internal_bestblockheight_get(self):
        """ Get the best stored block height. """

    @abc.abstractmethod
    def _internal_bestblockheight_put(self, height: int):
        """ Persist a new best stored block height. """

    @abc.abstractmethod
    def _internal_bestblockheight_update(self, height: int):
        """ Update the existing best stored block height. """

    @abc.abstractmethod
    def _internal_block_put(self, block: payloads.Block) -> None:
        """ Persist a block to the real backend. """

    @abc.abstractmethod
    def _internal_block_update(self, block: payloads.Block) -> None:
        """ Update a block in the real backend. """

    @abc.abstractmethod
    def _internal_block_delete(self, hash: types.UInt256) -> None:
        """ Delete a block from the real backend."""

    @abc.abstractmethod
    def _internal_block_get(self, hash: types.UInt256) -> payloads.Block:
        """
        Get a block from the real backend.

        Must raise KeyError if not found. Return value must be read only.
        """

    def _internal_block_try_get(self, hash: types.UInt256) -> Optional[payloads.Block]:
        try:
            return self._internal_block_get(hash)
        except KeyError:
            return None

    @abc.abstractmethod
    def _internal_block_get_by_height(self, height: int) -> payloads.Block:
        """
        Try to get a block by height.

        Must raise KeyError if not found.
        """

    @abc.abstractmethod
    def _internal_block_all(self) -> Iterator[payloads.Block]:
        """ Return all blocks stored in the real backend (readonly). """

    @abc.abstractmethod
    def _internal_contract_put(self, contract: contracts.ContractState) -> None:
        """ Persist a contract in the real backend. """

    @abc.abstractmethod
    def _internal_contract_update(self, contract: contracts.ContractState) -> None:
        """ Update a contract in the real backend. """

    @abc.abstractmethod
    def _internal_contract_delete(self, hash: types.UInt160) -> None:
        """ Delete a contract from the real backend. """

    @abc.abstractmethod
    def _internal_contract_get(self, hash: types.UInt160) -> contracts.ContractState:
        """
        Get a contract from the real backend.
        Must raise KeyError if not found. Return value must be read only.
        """

    def _internal_contract_try_get(self, hash: types.UInt160) -> Optional[contracts.ContractState]:
        try:
            return self._internal_contract_get(hash)
        except KeyError:
            return None

    @abc.abstractmethod
    def _internal_contract_all(self) -> Iterator[contracts.ContractState]:
        """ Return all contracts stored in the real backend (readonly). """

    @abc.abstractmethod
    def _internal_storage_put(self, key: storage.StorageKey, value: storage.StorageItem) -> None:
        """ Persist a storage key/value pair in the real backend. """

    @abc.abstractmethod
    def _internal_storage_update(self, key: storage.StorageKey, value: storage.StorageItem) -> None:
        """ Update a storage key/value pair in the real backend. """

    @abc.abstractmethod
    def _internal_storage_delete(self, key: storage.StorageKey) -> None:
        """ Delete a storage key/value pair from the real backend. """

    @abc.abstractmethod
    def _internal_storage_get(self, key: storage.StorageKey) -> storage.StorageItem:
        """
        Get a value by its key from the backend.

        Must raise KeyError if not found. Return value must be read only.
        """

    def _internal_storage_try_get(self, key: storage.StorageKey) -> Optional[storage.StorageItem]:
        try:
            return self._internal_storage_get(key)
        except KeyError:
            return None

    @abc.abstractmethod
    def _internal_storage_all(self, contract_id: Optional[int] = None) -> Iterator[Tuple[storage.StorageKey,
                                                                                         storage.StorageItem]]:
        """ Return all storage pairs for a given smart contract stored in the real backend (readonly). """

    @abc.abstractmethod
    def _internal_storage_find(self, key_prefix: bytes) -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        """ Find the storage pairs where the key property of the deserialized StorageKey starts with key_prefix. """

    def _internal_storage_seek(self,
                               key_prefix: bytes,
                               seek_direction="forward") -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        """ """

    @abc.abstractmethod
    def _internal_transaction_put(self, transaction: payloads.Transaction) -> None:
        """ Persist a transaction in the real backend (readonly). """

    @abc.abstractmethod
    def _internal_transaction_update(self, transaction: payloads.Transaction) -> None:
        """ Update a transaction in the real backend. """

    @abc.abstractmethod
    def _internal_transaction_delete(self, hash: types.UInt256) -> None:
        """ Delete a transaction from the real backend. """

    @abc.abstractmethod
    def _internal_transaction_get(self, hash: types.UInt256) -> payloads.Transaction:
        """
        Get a transaction from the real backend.

        Must raise KeyError if not found. Return value must be read only.
        """

    def _internal_transaction_try_get(self, hash: types.UInt256) -> Optional[payloads.Transaction]:
        try:
            return self._internal_transaction_get(hash)
        except KeyError:
            return None

    @abc.abstractmethod
    def _internal_transaction_all(self) -> Iterator[payloads.Transaction]:
        """ Return all transactions stored in the real backend. """


class RawView:
    def __init__(self, db: IDBImplementation):
        self._db = db

    @property
    def blocks(self):
        return RawBlockAccess(self._db)

    @property
    def contracts(self):
        return RawContractAccess(self._db)

    @property
    def storages(self):
        return RawStorageAccess(self._db)

    @property
    def transactions(self):
        return RawTXAccess(self._db)

    @property
    def best_block_height(self):
        try:
            return self._db._internal_bestblockheight_get()
        except KeyError:
            return -1

    @best_block_height.setter
    def best_block_height(self, value):
        raise AttributeError("Can't set attribute on a raw view. view.blocks.put() automatically updates the height "
                             "when applicable")


class RawBlockAccess:
    def __init__(self, db: IDBImplementation):
        self._db = db

    def put(self, block: payloads.Block) -> None:
        """
        Store a block.

        Args:
            block: instance.
        """
        self._db._internal_block_put(block)

    def get(self, hash: types.UInt256) -> payloads.Block:
        """
        Retrieve a block.

        Args:
            hash: block hash.

        Raises:
            KeyError: if the item is not found.
        """
        return self._db._internal_block_get(hash)

    def try_get(self, hash: types.UInt256) -> Optional[payloads.Block]:
        """
        Try to retrieve a block.

        Args:
            hash: block hash.
        """
        return self._db._internal_block_try_get(hash)

    def get_by_height(self, height: int) -> payloads.Block:
        """
        Retrieve a block by its height.

        Args:
            height: target block index/height.
        """
        return self._db._internal_block_get_by_height(height)

    def try_get_by_height(self, height: int) -> Optional[payloads.Block]:
        """
        Try to retrieve a block by its height.

        Args:
            height: target block index/height.
        """
        try:
            return self.get_by_height(height)
        except KeyError:
            return None

    def delete(self, hash: types.UInt256) -> None:
        """
        Remove a block.

        Args:
            hash: block hash.
        """
        self._db._internal_block_delete(hash)

    def all(self) -> Iterator[payloads.Block]:
        """
        Retrieve all stored blocks.
        """
        for block in self._db._internal_block_all():
            yield block


class RawContractAccess:
    def __init__(self, db: IDBImplementation):
        self._db = db

    def put(self, contract: contracts.ContractState) -> None:
        """
        Store a contract.
        Args:
            contract: contract state instance.
        """
        self._db._internal_contract_put(contract)

    def get(self, hash: types.UInt160) -> contracts.ContractState:
        """
        Retrieve a contract.
        Args:
            hash: unique contract identifier.
        Raises:
            KeyError: if the item is not found.
        """
        return self._db._internal_contract_get(hash)

    def try_get(self, hash: types.UInt160) -> Optional[contracts.ContractState]:
        """
        Try to retrieve a contract.
        Args:
            hash: unique contract identifier.
        """
        return self._db._internal_contract_try_get(hash)

    def delete(self, hash: types.UInt160) -> None:
        """
        Remove a contract.
        Args:
            hash: unique contract identifier.
        """
        self._db._internal_contract_delete(hash)

    def all(self) -> Iterator[contracts.ContractState]:
        """
        Retrieve all stored contracts.
        """
        for contract in self._db._internal_contract_all():
            yield contract


class RawStorageAccess:
    def __init__(self, db: IDBImplementation):
        self._db = db

    def put(self, key: storage.StorageKey, value: storage.StorageItem) -> None:
        """
        Store the value under the given key.

        Args:
            key: identifier to store the value under.
            value: the value to be persisted.
        """
        self._db._internal_storage_put(key, value)

    def get(self, key: storage.StorageKey) -> storage.StorageItem:
        """
        Retrieve a value from storage.

        Args:
            key: identifier.

        Raises:
            KeyError: if the item is not found.
        """
        return self._db._internal_storage_get(key)

    def try_get(self, key: storage.StorageKey) -> Optional[storage.StorageItem]:
        """
        Try to retrieve a value from storage.

        Args:
            key: identifier.
        """
        return self._db._internal_storage_try_get(key)

    def delete(self, key: storage.StorageKey) -> None:
        """
        Remove a key/value pair from storage

        Args:
            key: identifier to locate value.
        """
        self._db._internal_storage_delete(key)

    def all(self, contract_id: Optional[int] = None) -> Iterator[Tuple[storage.StorageKey,
                                                                       storage.StorageItem]]:
        """
        Retrieve all storage key/value pairs.

        Args:
            contract_id: smart contract unique identifier to limit results to. If not specified, returns for all
            contracts.
        """
        for k, v in self._db._internal_storage_all(contract_id):
            yield k, v

    def find(self, key_prefix: bytes) -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        """
        Find the storage pairs where the key property of the deserialized StorageKey starts with key_prefix.

        Args:
            key_prefix: the prefix part of the storage.StorageKey.key to look for.

        """
        for k, v in self._db._internal_storage_find(key_prefix):
            yield k, v


class RawTXAccess:
    def __init__(self, db: IDBImplementation):
        self._db = db

    def put(self, tx: payloads.Transaction) -> None:
        """
        Store a transaction.

        Args:
            tx: instance.
        """
        self._db._internal_transaction_put(tx)

    def get(self, hash: types.UInt256) -> payloads.Transaction:
        """
        Retrieve a transaction.

        Args:
            hash: transaction hash.

        Raises:
            KeyError: if the item is not found.
        """
        return self._db._internal_transaction_get(hash)

    def try_get(self, hash: types.UInt256) -> Optional[payloads.Transaction]:
        """
        Try to retrieve a transaction.

        Args:
            hash: transaction hash.
        """
        return self._db._internal_transaction_try_get(hash)

    def delete(self, hash: types.UInt256) -> None:
        """
        Remove a transaction.

        Args:
            hash: transaction hash.
        """
        self._db._internal_transaction_delete(hash)

    def all(self) -> Iterator[payloads.Transaction]:
        """
        Retrieve all stored transactions.
        """
        for tx in self._db._internal_transaction_all():
            yield tx


class StorageContext:
    def __init__(self, contract_id: int, is_read_only: bool):
        self.id = contract_id
        self.is_read_only = is_read_only



from __future__ import annotations
import hashlib
import struct
from typing import List
from neo3 import vm, storage, settings
from neo3.core import Size as s, serialization, types, utils, cryptography as crypto, IClonable, IInteroperable
from neo3.network import payloads
from bitarray import bitarray  # type: ignore
from copy import deepcopy
from .verification import IVerifiable


class Header(IVerifiable):
    """
    A Block header only object.

    Does not contain any consensus data or transactions.

    See also:
        :class:`~neo3.network.payloads.block.TrimmedBlock`
    """
    def __init__(self, version: int, prev_hash: types.UInt256, timestamp: int, nonce: int, index: int,
                 primary_index: int, next_consensus: types.UInt160, witness: payloads.Witness,
                 merkle_root: types.UInt256 = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.version = version
        self.prev_hash = prev_hash
        self.merkle_root = merkle_root if merkle_root else types.UInt256.zero()
        self.timestamp = timestamp
        self.nonce = nonce
        self.index = index
        self.primary_index = primary_index
        self.next_consensus = next_consensus
        self.witness = witness

    def __len__(self):
        return (s.uint32 + s.uint256 + s.uint256 + s.uint64 + s.uint64 + s.uint32 + s.uint8 + s.uint160 + 1
                + len(self.witness))

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    def hash(self) -> types.UInt256:
        """
        Get a unique identifier based on the unsigned data portion of the object.
        """
        with serialization.BinaryWriter() as bw:
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(data_to_hash).digest()
            return types.UInt256(data=data)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_serializable_list([self.witness])

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint32(self.version)
        writer.write_serializable(self.prev_hash)
        writer.write_serializable(self.merkle_root)
        writer.write_uint64(self.timestamp)
        writer.write_uint64(self.nonce)
        writer.write_uint32(self.index)
        writer.write_uint8(self.primary_index)
        writer.write_serializable(self.next_consensus)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the check byte does not equal.
        """
        self.deserialize_unsigned(reader)
        witnesses = reader.read_serializable_list(payloads.Witness, max=1)
        if len(witnesses) != 1:
            raise ValueError(f"Deserialization error - Witness object count is {len(witnesses)} must be 1")
        self.witness = witnesses[0]

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the unsigned data part of the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the primary_index field is greater than the configured consensus validator count.
        """
        (self.version,
         prev_hash,
         merkleroot,
         self.timestamp,
         self.nonce,
         self.index,
         self.primary_index,
         consensus) = struct.unpack("<I32s32sQQIB20s", reader._stream.read(109))
        if self.primary_index >= len(settings.standby_validators):
            raise ValueError(f"Deserialization error - primary index {self.primary_index} exceeds validator count "
                             f"{len(settings.standby_validators)}")
        self.prev_hash = types.UInt256(prev_hash)
        self.merkle_root = types.UInt256(merkleroot)
        self.next_consensus = types.UInt160(consensus)

    def get_script_hashes_for_verifying(self, snapshot: storage.Snapshot) -> List[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """
        if self.prev_hash == types.UInt256.zero():
            return [self.witness.script_hash()]
        prev_block = snapshot.blocks.try_get(self.prev_hash, read_only=True)
        if prev_block is None:
            raise ValueError("Can't get next_consensus hash from previous block. Block does not exist")
        return [prev_block.next_consensus]

    @classmethod
    def _serializable_init(cls):
        return cls(0,
                   types.UInt256.zero(),
                   0,
                   0,
                   0,
                   0,
                   types.UInt160.zero(),
                   payloads.Witness(b'', b''))


class Block(payloads.IInventory):
    """
    The famous Block. I transfer chain state.
    """

    def __init__(self,
                 header: Header,
                 transactions: List[payloads.Transaction] = None,
                 *args,
                 **kwargs
                 ):
        super(Block, self).__init__(*args, **kwargs)
        self.header = header
        self.transactions = [] if transactions is None else transactions

    def __len__(self):
        # calculate the varint length that needs to be inserted before the transaction objects.
        magic_len = utils.get_var_size(len(self.transactions))
        txs_len = sum([len(t) for t in self.transactions])
        return len(self.header) + magic_len + txs_len

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    @property
    def version(self) -> int:
        """ Block data structure version - for internal use """
        return self.header.version

    @property
    def prev_hash(self) -> types.UInt256:
        """ The hash of the previous block """
        return self.header.prev_hash

    @property
    def merkle_root(self) -> types.UInt256:
        """ The merkle root of the transactions in the block """
        return self.header.merkle_root

    @property
    def timestamp(self) -> int:
        """ UTC timestamp in miliseconds """
        return self.header.timestamp

    @property
    def nonce(self) -> int:
        """ Random number """
        return self.header.nonce

    @property
    def index(self) -> int:
        """ The height of the block """
        return self.header.index

    @property
    def primary_index(self) -> int:
        """ The index into the consensus node list that was used to generate this block """
        return self.header.primary_index

    @property
    def next_consensus(self) -> types.UInt160:
        """ The hash of the consensus node that will generate the next block """
        return self.header.next_consensus

    @property
    def witness(self) -> payloads.Witness:
        """ The witness of this block """
        return self.header.witness

    def hash(self) -> types.UInt256:
        """ A unique identifier based on the unsigned data portion of the object """
        return self.header.hash()

    def get_script_hashes_for_verifying(self, snapshot: storage.Snapshot) -> List[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """
        return self.header.get_script_hashes_for_verifying(snapshot)

    @property
    def inventory_type(self) -> payloads.InventoryType:
        """
        Inventory type identifier.
        """
        return payloads.InventoryType.BLOCK

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.header)
        writer.write_var_int(len(self.transactions))
        for tx in self.transactions:
            writer.write_serializable(tx)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the unsigned part of the object into a binary stream.

        Args:
            writer: instance.
        """
        self.header.serialize_unsigned(writer)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the content count of the block is zero, or if there is a duplicate transaction in the list,
                or if the merkle root does not included the calculated root.
        """
        self.header = reader.read_serializable(Header)
        self.transactions = reader.read_serializable_list(payloads.Transaction, max=0xFFFF)

        if len(set(self.transactions)) != len(self.transactions):
            raise ValueError("Deserialization error - block contains duplicate transaction")

        hashes = [t.hash() for t in self.transactions]
        if crypto.MerkleTree.compute_root(hashes) != self.header.merkle_root:
            raise ValueError("Deserialization error - merkle root mismatch")

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """ Not supported """
        raise NotImplementedError

    def rebuild_merkle_root(self) -> None:
        """
        Recalculates the Merkle root.
        """
        self.header.merkle_root = crypto.MerkleTree.compute_root([t.hash() for t in self.transactions])

    def trim(self) -> TrimmedBlock:
        """
        Reduce a block in size by replacing the transaction objects with their identifying hashes.
        """
        return TrimmedBlock(self.header, [t.hash() for t in self.transactions])

    def from_replica(self, replica: Block) -> None:
        """
        Shallow copy attributes from a reference object.
        """
        self.header = replica.header
        self.transactions = replica.transactions

    @classmethod
    def _serializable_init(cls):
        return cls(Header._serializable_init(), [])


class TrimmedBlock(serialization.ISerializable, IInteroperable):
    """
    A size reduced Block instance.

    Contains consensus data and transactions hashes instead of their full objects.
    """

    def __init__(self, header: Header, hashes: List[types.UInt256]):
        super(TrimmedBlock, self).__init__()
        self.header = header
        self.hashes = hashes

    def __len__(self):
        return len(self.header) + utils.get_var_size(self.hashes)

    def __deepcopy__(self, memodict={}):
        # not the best, but faster than letting deepcopy() do introspection
        return self.__class__.deserialize_from_bytes(self.to_array())

    def hash(self):
        """ A unique identifier based on the unsigned data portion of the object """
        return self.header.hash()

    @property
    def index(self):
        """ The height of the block """
        return self.header.index

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.header)
        writer.write_serializable_list(self.hashes)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.header = reader.read_serializable(Header)
        self.hashes = reader.read_serializable_list(types.UInt256, max=0xFFFF)

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        array = vm.ArrayStackItem(reference_counter)
        array.append(vm.ByteStringStackItem(self.header.hash().to_array()))
        array.append(vm.IntegerStackItem(self.header.version))
        array.append(vm.ByteStringStackItem(self.header.prev_hash.to_array()))
        array.append(vm.ByteStringStackItem(self.header.merkle_root.to_array()))
        array.append(vm.IntegerStackItem(self.header.timestamp))
        array.append(vm.IntegerStackItem(self.header.nonce))
        array.append(vm.IntegerStackItem(self.header.index))
        array.append(vm.IntegerStackItem(self.header.primary_index))
        array.append(vm.ByteStringStackItem(self.header.next_consensus.to_array()))
        array.append(vm.IntegerStackItem(len(self.hashes)))
        return array

    @classmethod
    def _serializable_init(cls):
        return cls(Header._serializable_init(), [])


class MerkleBlockPayload(serialization.ISerializable):
    def __init__(self, block: Block, flags: bitarray):
        hashes = [t.hash() for t in block.transactions]
        tree = crypto.MerkleTree(hashes)
        self.flags = flags.tobytes()
        self.tx_count = len(hashes)
        self.hashes = tree.to_hash_array()
        self.header = block.header

    def __len__(self):
        return len(self.header) + s.uint32 + utils.get_var_size(self.hashes) + utils.get_var_size(self.flags)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.header)
        writer.write_var_int(self.tx_count)
        writer.write_serializable_list(self.hashes)
        writer.write_var_bytes(self.flags)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.header = reader.read_serializable(Header)
        self.tx_count = reader.read_var_int(max=0xFFFF)
        self.hashes = reader.read_serializable_list(types.UInt256, max=self.tx_count)
        self.flags = reader.read_var_bytes(max=(max(self.tx_count, 1) + 7) // 8)

    @classmethod
    def _serializable_init(cls):
        block = payloads.Block._serializable_init()
        flags = bitarray()
        return cls(block, flags)


class HeadersPayload(serialization.ISerializable):
    MAX_HEADERS_COUNT = 2000

    def __init__(self, headers: List[Header] = None):
        """
        Should not be called directly. Use create() instead.
        """
        self.headers = headers if headers else []

    def __len__(self):
        return utils.get_var_size(self.headers)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable_list(self.headers)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.headers = reader.read_serializable_list(Header)


class GetBlocksPayload(serialization.ISerializable):
    """
    Used to request an array block hashes that can be retrieved via a message with the
    :const:`~neo3.network.message.MessageType.GETDATA` type.
    """
    def __init__(self, hash_start: types.UInt256, count=-1):
        """
        Create payload.

        Args:
            hash_start: starting point from which to return the `next` hash.

                Note:

                    For syncing supply the local best height block hash to receive the hashes in the range of
                    best_height+1 to best_height+1+count

            count: number of hashes to return.
        """
        self.hash_start = hash_start
        self.count = count

    def __len__(self):
        return s.uint256 + s.uint16

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.hash_start)
        writer.write_int16(self.count)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.hash_start = reader.read_serializable(types.UInt256)
        self.count = reader.read_int16()

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt256.zero())


class GetBlockByIndexPayload(serialization.ISerializable):
    """
    Used to request full Block or Header objects via a message with the
    :const:`~neo3.network.message.MessageType.GETBLOCKBYINDEX` or :const:`~neo3.network.message.MessageType.GETHEADERS`
    type respectively.
    """

    def __init__(self, index_start: int, count: int = HeadersPayload.MAX_HEADERS_COUNT):
        """
        Create payload.

        Args:
            index_start: start block height.
            count: number of blocks or headers to requests starting from `index_start`.

        """
        self.index_start = index_start
        self.count = count

    def __len__(self):
        return s.uint32 + s.uint16

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.index_start)
        writer.write_int16(self.count)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if `count` is zero or exceeds
               :const:`~neo3.network.payloads.getblocks.GetBlockByIndexPayload.MAX_BLOCKS_COUNT`.
        """
        self.index_start = reader.read_uint32()
        self.count = reader.read_int16()
        if self.count < 1 or self.count == 0 or self.count > HeadersPayload.MAX_HEADERS_COUNT:
            raise ValueError("Deserialization error - invalid count")

    @classmethod
    def _serializable_init(cls):
        return cls(0)



from __future__ import annotations

import base64
import hashlib
from typing import List
from neo3.core import serialization, types, Size as s, utils, IJson
from neo3 import contracts


class NEF(serialization.ISerializable, IJson):
    def __init__(self,
                 compiler_name: str = None,
                 script: bytes = None,
                 tokens: List[MethodToken] = None,
                 source: str = None,
                 _magic: int = 0x3346454E):
        """
        Create a Neo Executable Format file.

        Args:
            compiler_name: human readable name of the compiler and version used to create the script data.
            Automatically limited to 64 bytes
            script: a byte array of raw VM opcodes.
        """
        self.magic = _magic
        if compiler_name is None:
            self.compiler = 'unknown'
        else:
            self.compiler = compiler_name[:64]
        self.source = source if source else ""
        self.script = script if script else b''
        self._checksum = 0
        self.tokens = [] if tokens is None else tokens
        # this is intentional, because NEO computes the initial checksum by serializing itself while checksum is 0
        self._checksum = self.compute_checksum()

    def __len__(self):
        return (
            s.uint32  # magic
            + 64  # compiler
            + utils.get_var_size(self.source)
            + 1  # reserved
            + utils.get_var_size(self.tokens)
            + 2  # reserved
            + s.uint32  # checksum
            + utils.get_var_size(self.script))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.magic == other.magic
                and self.compiler == other.compiler
                and self.source == other.source
                and self.script == other.script
                and self.tokens == other.tokens
                and self.checksum == other.checksum)

    @property
    def checksum(self) -> int:
        if self._checksum == 0:
            self._checksum = self.compute_checksum()
        return self._checksum

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.magic)
        writer.write_bytes(self.compiler.encode('utf-8').ljust(64, b'\x00'))
        writer.write_var_string(self.source)
        writer.write_bytes(b'\x00')
        writer.write_serializable_list(self.tokens)
        writer.write_bytes(b'\x00\x00')
        writer.write_var_bytes(self.script)
        writer.write_uint32(self._checksum)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        if reader.read_uint32() != self.magic:
            raise ValueError("Deserialization error - Incorrect magic")
        self.compiler = reader.read_bytes(64).decode('utf-8').rstrip(b'\x00'.decode())
        self.source = reader.read_var_string(256)
        if reader.read_uint8() != 0:
            raise ValueError("Reserved bytes must be 0")
        self.tokens = reader.read_serializable_list(MethodToken)
        if reader.read_uint16() != 0:
            raise ValueError("Reserved bytes must be 0")

        self.script = reader.read_var_bytes(max=512 * 1024)
        if len(self.script) == 0:
            raise ValueError("Deserialization error - Script can't be empty")

        checksum = int.from_bytes(reader.read_bytes(4), 'little')
        if checksum != self.compute_checksum():
            raise ValueError("Deserialization error - Invalid checksum")
        else:
            self._checksum = checksum

    def compute_checksum(self) -> int:
        """
        Compute the checksum of the NEF file.
        """
        return int.from_bytes(hashlib.sha256(hashlib.sha256(self.to_array()[:-4]).digest()).digest()[:4], 'little')

    def to_json(self) -> dict:
        raise NotImplementedError

    @classmethod
    def from_json(cls, json: dict):
        tokens = []
        for t in json['tokens']:
            tokens.append(MethodToken.from_json(t))
        script = base64.b64decode(json['script'])
        return cls(json['compiler'], script, tokens, _magic=json['magic'])

    @classmethod
    def _serializable_init(cls):
        c = cls()
        c._checksum = 0
        return c


class MethodToken(serialization.ISerializable, IJson):
    def __init__(self,
                 hash: types.UInt160,
                 method: str,
                 parameters_count: int,
                 has_return_value: bool,
                 call_flags: contracts.CallFlags):
        self.hash = hash
        self.method = method
        self.parameters_count = parameters_count
        self.has_return_value = has_return_value
        self.call_flags = call_flags

    def __len__(self):
        return s.uint160 + utils.get_var_size(self.method) + s.uint16 + s.uint8 + s.uint8

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.hash)
        writer.write_var_string(self.method)
        writer.write_uint16(self.parameters_count)
        writer.write_uint8(self.has_return_value)
        writer.write_uint8(self.call_flags.value)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.hash = reader.read_serializable(types.UInt160)
        self.method = reader.read_var_string(32)
        self.parameters_count = reader.read_uint16()
        self.has_return_value = bool(reader.read_uint8())
        self.call_flags = contracts.CallFlags(reader.read_uint8())

    def to_json(self) -> dict:
        raise NotImplementedError

    @classmethod
    def from_json(cls, json: dict):
        h = types.UInt160.from_string(json['hash'][2:])
        m = json['method']
        pc = json['paramcount']
        rv = json['hasreturnvalue']
        cf = contracts.CallFlags.from_csharp_name(json['callflags'])
        return cls(h, m, pc, rv, cf)

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt160.zero(), "", 0, False, contracts.CallFlags.NONE)



import unittest
from unittest import mock
import binascii
from neo3 import settings, contracts, storage, vm
from neo3.core import cryptography, types, syscall_name_to_int, to_script_hash, msgrouter
from tests.contracts.interop.utils import test_engine, test_block, TestIVerifiable, contract_hash


class Nep5TestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        settings.network.standby_committee = ['02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765']
        settings.network.validators_count = 1
        cls.validator_public_key = cryptography.ECPoint.deserialize_from_bytes(
            binascii.unhexlify(settings.network.standby_committee[0])
        )
        cls.validator_account = to_script_hash(
            contracts.Contract.create_multisig_redeemscript(1, [cls.validator_public_key]))

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()

    def test_token_symbols(self):
        gas_symbol = contracts.GasToken().symbol()
        neo_symbol = contracts.NeoToken().symbol()
        self.assertEqual("GAS", gas_symbol)
        self.assertEqual("NEO", neo_symbol)

    def test_total_supply(self):
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        # set or we won't pass the native deploy call
        engine.snapshot.persisting_block = block

        gas = contracts.GasToken()
        neo = contracts.NeoToken()
        # this is now the initial stored value + the result of the post_persist event (=committee reward added)
        expected_total_gas_supply = vm.BigInteger(5200000050000000)
        self.assertEqual(expected_total_gas_supply, gas.total_supply(engine.snapshot))
        self.assertEqual(100_000_000, neo.total_supply(engine.snapshot))

    def test_burn(self):
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        # set or we won't pass the native deploy call
        engine.snapshot.persisting_block = block

        gas = contracts.GasToken()

        with self.assertRaises(ValueError) as context:
            gas.burn(engine, self.validator_account, vm.BigInteger(-1))
        self.assertEqual("Can't burn a negative amount", str(context.exception))

        default_gas = 52_000_000
        self.assertEqual(default_gas, gas.balance_of(engine.snapshot, self.validator_account) / gas.factor)
        gas.burn(engine, self.validator_account, vm.BigInteger(0))
        self.assertEqual(default_gas, gas.balance_of(engine.snapshot, self.validator_account) / gas.factor)

        with self.assertRaises(ValueError) as context:
            gas.burn(engine, self.validator_account, vm.BigInteger(default_gas + 1) * gas.factor)
        self.assertEqual("Insufficient balance. Requesting to burn 5200000100000000, available 5200000000000000",
                         str(context.exception))

        # burn a bit
        gas.burn(engine, self.validator_account, vm.BigInteger(10) * gas.factor)
        remaining_balance = int(gas.balance_of(engine.snapshot, self.validator_account) / gas.factor)
        self.assertEqual(default_gas - 10, remaining_balance)

        # now burn it all
        gas.burn(engine, self.validator_account, vm.BigInteger(remaining_balance) * gas.factor)
        self.assertEqual(0, gas.balance_of(engine.snapshot, self.validator_account) / gas.factor)

    def test_balance_of(self):
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        # set or we won't pass the native deploy call
        engine.snapshot.persisting_block = block

        gas = contracts.GasToken()
        neo = contracts.NeoToken()

        deploy_expected_gas = 52_000_000
        deploy_expected_neo = 100_000_000
        self.assertEqual(deploy_expected_gas, gas.balance_of(engine.snapshot, self.validator_account) / gas.factor)
        self.assertEqual(deploy_expected_neo, neo.balance_of(engine.snapshot, self.validator_account))

        self.assertEqual(vm.BigInteger.zero(), gas.balance_of(engine.snapshot, types.UInt160.zero()))
        self.assertEqual(vm.BigInteger.zero(), neo.balance_of(engine.snapshot, types.UInt160.zero()))

    def test_on_persist(self):
        """
        OnPersist will do the following
        * burn the system and network fees for all transactions
        * mint the sum of network_fees for all transactions to the address of the consensus node that acted as primary
          speaker for the block
        """
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        # set or we won't pass the native deploy call
        engine.snapshot.persisting_block = block

        gas = contracts.GasToken()

        # update the TX signer account to point to our validator or the token burn() (part of on persist)
        # will fail because it can't find an account with balance
        mock_signer = mock.MagicMock()
        mock_signer.account = self.validator_account
        engine.snapshot.persisting_block.transactions[0].signers = [mock_signer]
        # our consensus_data is not setup in a realistic way, so we have to correct for that here
        # or we fail to get the account of primary consensus node
        engine.snapshot.persisting_block.header.primary_index = settings.network.validators_count - 1

        gas.on_persist(engine)

        """
            Drop the below in a test in UT_NativeContract.cs and change ProtocolSettings.cs to
            * have a ValidatorsCount of 1 
            * and the StandbyCommittee should be: 02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765

            var snapshot = Blockchain.Singleton.GetSnapshot();
            snapshot.PersistingBlock = new Block() { Index = 1000 };
            var point = ECPoint.Parse("02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765", ECCurve.Secp256r1);
            var account = Contract.CreateMultiSigRedeemScript(1, new ECPoint[] {point}).ToScriptHash();
            var tx = TestUtils.GetTransaction(account);
            tx.SystemFee = 456;
            tx.NetworkFee = 789;
            snapshot.PersistingBlock.Transactions = new Transaction[] {tx};
            snapshot.PersistingBlock.ConsensusData = new ConsensusData { PrimaryIndex = 0};

            ApplicationEngine engine2 = ApplicationEngine.Create(TriggerType.System, tx, snapshot, 0);
            NativeContract.GAS.OnPersist(engine2);
            var key = new byte[] {0x14};
            var sk  = key.Concat(account.ToArray());
            var item = engine2.Snapshot.Storages.TryGet(new StorageKey {Id = NativeContract.GAS.Id, Key = sk.ToArray()});
            var state = item.GetInteroperable<AccountState>();
            Console.WriteLine($"account state {state.Balance}");

            var item2 = engine2.Snapshot.Storages.TryGet(new StorageKey {Id = NativeContract.GAS.Id, Key = new byte[]{11}});
            Console.WriteLine($"total supply {(BigInteger)item2}");

            var primary_account = Contract.CreateSignatureRedeemScript(point).ToScriptHash();
            var primary_sk = key.Concat(primary_account.ToArray());
            var primary_item = engine2.Snapshot.Storages.TryGet(new StorageKey {Id = NativeContract.GAS.Id, Key = primary_sk.ToArray()});
            var primary_state = primary_item.GetInteroperable<AccountState>();
            Console.WriteLine($"primary account state {primary_state.Balance}");
        """

        # * our validator prior to on_persist had a balance of 30_000_000
        # * after it should have been reduced by the network + system_fee's paid in the transaction
        sk_gas_supply = gas.key_account + self.validator_account
        si_supply = engine.snapshot.storages.try_get(sk_gas_supply)
        self.assertIsNotNone(si_supply)
        token_state = gas._state.deserialize_from_bytes(si_supply.value)
        total_fees = engine.snapshot.persisting_block.transactions[0].network_fee + \
                     engine.snapshot.persisting_block.transactions[0].system_fee
        expected = (52_000_000 * gas.factor) - total_fees
        self.assertEqual(expected, int(token_state.balance))

        # * total GAS supply was 30_000_000 + 0.5 for committee reward, should be reduced by the system_fee
        sk_total_supply = gas.key_total_supply
        si_total_supply = engine.snapshot.storages.try_get(sk_total_supply)
        self.assertIsNotNone(si_total_supply)
        committee_reward = vm.BigInteger(50000000)
        expected = ((52_000_000 * gas.factor) + committee_reward) - engine.snapshot.persisting_block.transactions[0].system_fee
        self.assertEqual(expected, vm.BigInteger(si_total_supply.value))

        # * the persisting block contains exactly 1 transaction
        # * after on_persist the account our primary validator should have been credited with the transaction's
        #   network_fee
        primary_validator = to_script_hash(contracts.Contract.create_signature_redeemscript(self.validator_public_key))
        sk_gas_supply = gas.key_account + primary_validator
        si_supply = engine.snapshot.storages.try_get(sk_gas_supply)
        self.assertIsNotNone(si_supply)
        token_state = gas._state.deserialize_from_bytes(si_supply.value)
        expected = engine.snapshot.persisting_block.transactions[0].network_fee + committee_reward
        self.assertEqual(expected, int(token_state.balance))

    def transfer_helper(self, contract: contracts.NativeContract,
                        from_account: types.UInt160,
                        to_account: types.UInt160,
                        amount: vm.BigInteger):
        engine = test_engine(has_snapshot=True)
        block = test_block(0)
        # set or we won't pass the native deploy call
        engine.snapshot.persisting_block = block
        engine.invocation_stack.pop()  # we no longer need the default script
        engine.script_container = TestIVerifiable()
        engine.script_container.script_hashes = [from_account]

        sb = vm.ScriptBuilder()
        sb.emit(vm.OpCode.PUSHNULL)
        sb.emit_push(amount)
        sb.emit_push(to_account.to_array())
        sb.emit_push(from_account.to_array())
        sb.emit_push(4)
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(15)  # callflags
        sb.emit_push(b'transfer')
        sb.emit_push(contract.hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))
        engine.load_script(vm.Script(sb.to_array()))

        nef = contracts.NEF(script=sb.to_array())
        manifest = contracts.ContractManifest("test_contract")
        contract_state = contracts.ContractState(1, nef, manifest, 0, contract_hash(from_account, nef.checksum, manifest.name))
        engine.snapshot.contracts.put(contract_state)
        return engine

    def test_transfer_negative_amount(self):
        engine = test_engine(has_snapshot=True, default_script=False)
        engine.load_script(vm.Script(contracts.GasToken().script))
        block = test_block(0)
        # set or we won't pass the native deploy call
        engine.snapshot.persisting_block = block

        gas = contracts.GasToken()

        with self.assertRaises(ValueError) as context:
            gas.transfer(engine, types.UInt160.zero(), types.UInt160.zero(), vm.BigInteger(-1), vm.NullStackItem())
        self.assertEqual("Can't transfer a negative amount", str(context.exception))

    def test_transfer_fail_no_permission(self):
        """
        Test to transfer tokens from a source account that is not owned by the smart contract that is asking for the transfer.
        We do not add a witness that approves that we can transfer from the source account, thus it should fail
        Returns:

        """
        gas = contracts.GasToken()

        engine = self.transfer_helper(gas, types.UInt160.zero(), types.UInt160.zero(), vm.BigInteger(1))
        engine.script_container.script_hashes = []  # ensure checkwitness returns False
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        result = engine.result_stack.pop()
        self.assertFalse(result)

    def test_transfer_from_empty_account(self):
        gas = contracts.GasToken()
        manifest = contracts.ContractManifest("test_contract2")
        nef = contracts.NEF(script=b'\x40')
        state = contracts.ContractState(2, nef, manifest, 0, contract_hash(types.UInt160.zero(), nef.checksum, manifest.name))

        engine = self.transfer_helper(gas, types.UInt160.zero(), state.hash, vm.BigInteger(1))
        engine.snapshot.contracts.put(state)
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        result = engine.result_stack.pop()
        self.assertFalse(result)

    def test_transfer_zero_amount(self):
        gas = contracts.GasToken()
        account_from = types.UInt160(b'\x01' * 20)
        storage_key_from = gas.key_account + account_from
        account_state = gas._state()
        account_state.balance = vm.BigInteger(123)
        storage_item_from = storage.StorageItem(account_state.to_array())

        manifest_to = contracts.ContractManifest("source_contract")
        nef_to = contracts.NEF(script=b'\x40')
        state_to = contracts.ContractState(1, nef_to, manifest_to, 0, contract_hash(types.UInt160.zero(), nef_to.checksum, manifest_to.name))
        account_to = state_to.hash
        amount = vm.BigInteger(0)

        engine = self.transfer_helper(gas, account_from, account_to, amount)

        # ensure the source account has balance
        engine.snapshot.storages.put(storage_key_from, storage_item_from)

        transfer_event = ()

        def notify_listener(contract_script_hash, event, state):
            nonlocal transfer_event
            transfer_event = (contract_script_hash, event, state)

        msgrouter.interop_notify += notify_listener
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        result = engine.result_stack.pop()
        self.assertTrue(result)
        self.assertEqual(gas.hash, transfer_event[0])
        self.assertEqual("Transfer", transfer_event[1])
        state_items = list(transfer_event[2])
        self.assertEqual(account_from, types.UInt160(state_items[0].to_array()))
        self.assertEqual(account_to, types.UInt160(state_items[1].to_array()))
        self.assertEqual(amount, state_items[2].to_biginteger())

    def test_transfer_more_than_balance(self):
        gas = contracts.GasToken()
        account_from = types.UInt160.zero()
        storage_key_from = gas.key_account + account_from
        account_state = gas._state()
        account_state.balance = vm.BigInteger(123)
        storage_item_from = storage.StorageItem(account_state.to_array())

        manifest_to = contracts.ContractManifest("source_contract")
        nef_to = contracts.NEF(script=b'\x40')
        state_to = contracts.ContractState(1, nef_to, manifest_to, 0, contract_hash(types.UInt160.zero(), nef_to.checksum, manifest_to.name))
        account_to = state_to.hash
        amount = account_state.balance + 1

        engine = self.transfer_helper(gas, account_from, account_to, amount)
        # ensure the source account has balance
        engine.snapshot.storages.put(storage_key_from, storage_item_from)
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        result = engine.result_stack.pop()
        self.assertFalse(result)

    def test_transfer_to_self(self):
        gas = contracts.GasToken()

        manifest = contracts.ContractManifest("test_contract")
        nef = contracts.NEF(script=b'\x40')
        state_to = contracts.ContractState(1, nef, manifest, 0, contract_hash(types.UInt160.zero(), nef.checksum, manifest.name))
        account = state_to.hash

        storage_key_from = gas.key_account + account
        account_state = gas._state()
        account_state.balance = vm.BigInteger(123)
        storage_item_from = storage.StorageItem(account_state.to_array())

        amount = account_state.balance

        engine = self.transfer_helper(gas, account, account, amount)
        # ensure the source account has balance
        engine.snapshot.storages.put(storage_key_from, storage_item_from)

        transfer_event = ()

        def notify_listener(contract_script_hash, event, state):
            nonlocal transfer_event
            transfer_event = (contract_script_hash, event, state)

        msgrouter.interop_notify += notify_listener
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        result = engine.result_stack.pop()
        self.assertTrue(result)
        self.assertEqual(gas.hash, transfer_event[0])
        self.assertEqual("Transfer", transfer_event[1])
        state_items = list(transfer_event[2])
        self.assertEqual(account, types.UInt160(state_items[0].to_array()))
        self.assertEqual(account, types.UInt160(state_items[1].to_array()))
        self.assertEqual(amount, state_items[2].to_biginteger())

    def test_transfer_full_balance(self):
        gas = contracts.GasToken()

        manifest = contracts.ContractManifest("contract_name_to")
        nef = contracts.NEF(script=b'\x40')
        state_to = contracts.ContractState(1, nef, manifest, 0, contract_hash(types.UInt160.zero(), nef.checksum, manifest.name))
        account_to = state_to.hash

        account_from = types.UInt160(b'\x01' * 20)
        storage_key_from = gas.key_account + account_from
        account_from_state = gas._state()
        account_from_state.balance = vm.BigInteger(123)
        storage_item_from = storage.StorageItem(account_from_state.to_array())

        amount = account_from_state.balance

        engine = self.transfer_helper(gas, account_from, account_to, amount)

        # ensure the source account has balance
        engine.snapshot.storages.put(storage_key_from, storage_item_from)

        transfer_event = ()

        def notify_listener(contract_script_hash, event, state):
            nonlocal transfer_event
            transfer_event = (contract_script_hash, event, state)

        msgrouter.interop_notify += notify_listener
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        result = engine.result_stack.pop()
        self.assertTrue(result)
        self.assertEqual(gas.hash, transfer_event[0])
        self.assertEqual("Transfer", transfer_event[1])
        state_items = list(transfer_event[2])
        self.assertEqual(account_from, types.UInt160(state_items[0].to_array()))
        self.assertEqual(account_to, types.UInt160(state_items[1].to_array()))
        self.assertEqual(amount, state_items[2].to_biginteger())

        # test that the source account is no longer present in storage as the balance is zero
        self.assertIsNone(engine.snapshot.storages.try_get(storage_key_from))

    def test_transfer_partial_balance_to_account_with_balance(self):
        gas = contracts.GasToken()

        manifest = contracts.ContractManifest("contract_name")
        nef = contracts.NEF(script=b'\x40')
        state_to = contracts.ContractState(1, nef, manifest, 0, contract_hash(types.UInt160.zero(), nef.checksum, manifest.name))
        account_to = state_to.hash
        storage_key_to = gas.key_account + account_to
        account_to_state = gas._state()
        account_to_state.balance = vm.BigInteger(100)
        storage_item_to = storage.StorageItem(account_to_state.to_array())

        account_from = types.UInt160(b'\x01' * 20)
        storage_key_from = gas.key_account + account_from
        account_from_state = gas._state()
        account_from_state.balance = vm.BigInteger(123)
        storage_item_from = storage.StorageItem(account_from_state.to_array())

        amount = vm.BigInteger(50)

        engine = self.transfer_helper(gas, account_from, account_to, amount)
        # ensure the source and destination account have balances
        engine.snapshot.storages.put(storage_key_from, storage_item_from)
        engine.snapshot.storages.put(storage_key_to, storage_item_to)

        transfer_event = ()

        def notify_listener(contract_script_hash, event, state):
            nonlocal transfer_event
            transfer_event = (contract_script_hash, event, state)

        msgrouter.interop_notify += notify_listener
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        result = engine.result_stack.pop()
        self.assertTrue(result)
        self.assertEqual(gas.hash, transfer_event[0])
        self.assertEqual("Transfer", transfer_event[1])
        state_items = list(transfer_event[2])
        self.assertEqual(account_from, types.UInt160(state_items[0].to_array()))
        self.assertEqual(account_to, types.UInt160(state_items[1].to_array()))
        self.assertEqual(amount, state_items[2].to_biginteger())

        # validate from account is deducted by `amount`
        new_storage_account_from = engine.snapshot.storages.get(storage_key_from)
        new_account_state_from = gas._state.deserialize_from_bytes(new_storage_account_from.value)
        self.assertEqual(account_from_state.balance - amount, new_account_state_from.balance)

        # validate to account is credited with `amount`
        new_storage_account_to = engine.snapshot.storages.get(storage_key_to)
        new_account_state_to = gas._state.deserialize_from_bytes(new_storage_account_to.value)
        self.assertEqual(account_to_state.balance + amount, new_account_state_to.balance)

    def test_negative_mint(self):
        gas = contracts.GasToken()
        with self.assertRaises(ValueError) as context:
            gas.mint(None, None, vm.BigInteger(-1), False)
        self.assertEqual("Can't mint a negative amount", str(context.exception))



def transform(app, doctree, fromdocname):
    if "bloomfilter" in fromdocname:
        print("hi")

def transform2(app, doctree):
    print("ok")

def setup(app):
    app.connect('doctree-resolved', transform)
    app.connect('doctree-read', transform2)


    # remove after debug
    # app.connect('builder-inited', add_jinja_filters)

    return {
        'version': '0.1',
        'parallel_read_safe': False,
        'parallel_write_safe': False,
    }


from neo3.core import serialization


class EmptyPayload(serialization.ISerializable):
    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """ we don't have to do anything, because it should stay empty. """

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """ we don't have to do anything, because it has no attributes. """

    def __len__(self):
        return 0



import unittest
from neo3.network.ipfilter import IPFilter


class IPFilteringTestCase(unittest.TestCase):
    def test_nobody_allowed(self):
        filter = IPFilter()
        filter.load_config({
            'blacklist': [
                '0.0.0.0/0'
            ],
            'whitelist': [
            ]
        })

        self.assertFalse(filter.is_allowed('127.0.0.1'))
        self.assertFalse(filter.is_allowed('10.10.10.10'))

    def test_nobody_allowed_except_one(self):
        filter = IPFilter()
        filter.load_config({
            'blacklist': [
                '0.0.0.0/0'
            ],
            'whitelist': [
                '10.10.10.10'
            ]
        })

        self.assertFalse(filter.is_allowed('127.0.0.1'))
        self.assertFalse(filter.is_allowed('10.10.10.11'))
        self.assertTrue(filter.is_allowed('10.10.10.10'))

    def test_everybody_allowed(self):
        filter = IPFilter()
        filter.load_config({
            'blacklist': [
            ],
            'whitelist': [
            ]
        })

        self.assertTrue(filter.is_allowed('127.0.0.1'))
        self.assertTrue(filter.is_allowed('10.10.10.11'))
        self.assertTrue(filter.is_allowed('10.10.10.10'))

        filter.load_config({
            'blacklist': [
            ],
            'whitelist': [
                '0.0.0.0/0'
            ]
        })

        self.assertTrue(filter.is_allowed('127.0.0.1'))
        self.assertTrue(filter.is_allowed('10.10.10.11'))
        self.assertTrue(filter.is_allowed('10.10.10.10'))

        filter.load_config({
            'blacklist': [
                '0.0.0.0/0'
            ],
            'whitelist': [
                '0.0.0.0/0'
            ]
        })

        self.assertTrue(filter.is_allowed('127.0.0.1'))
        self.assertTrue(filter.is_allowed('10.10.10.11'))
        self.assertTrue(filter.is_allowed('10.10.10.10'))

    def test_everybody_allowed_except_one(self):
        filter = IPFilter()
        filter.load_config({
            'blacklist': [
                '127.0.0.1'
            ],
            'whitelist': [
            ]
        })

        self.assertFalse(filter.is_allowed('127.0.0.1'))
        self.assertTrue(filter.is_allowed('10.10.10.11'))
        self.assertTrue(filter.is_allowed('10.10.10.10'))

    def test_disallow_ip_range(self):
        filter = IPFilter()
        filter.load_config({
            'blacklist': [
                '127.0.0.0/24'
            ],
            'whitelist': [
            ]
        })

        self.assertFalse(filter.is_allowed('127.0.0.0'))
        self.assertFalse(filter.is_allowed('127.0.0.1'))
        self.assertFalse(filter.is_allowed('127.0.0.100'))
        self.assertFalse(filter.is_allowed('127.0.0.255'))
        self.assertTrue(filter.is_allowed('10.10.10.11'))
        self.assertTrue(filter.is_allowed('10.10.10.10'))

    def test_updating_blacklist(self):
        filter = IPFilter()
        filter.load_config({
            'blacklist': [
            ],
            'whitelist': [
            ]
        })

        self.assertTrue(filter.is_allowed('127.0.0.1'))

        filter.blacklist_add('127.0.0.0/24')
        self.assertFalse(filter.is_allowed('127.0.0.1'))
        # should have no effect, only exact matches
        filter.blacklist_remove('127.0.0.1')
        self.assertFalse(filter.is_allowed('127.0.0.1'))

        filter.blacklist_remove('127.0.0.0/24')
        self.assertTrue(filter.is_allowed('127.0.0.1'))

    def test_updating_whitelist(self):
        filter = IPFilter()
        filter.load_config({
            'blacklist': [
                '0.0.0.0/0'
            ],
            'whitelist': [
            ]
        })

        self.assertFalse(filter.is_allowed('127.0.0.1'))

        filter.whitelist_add('127.0.0.0/24')
        self.assertTrue(filter.is_allowed('127.0.0.1'))

        filter.whitelist_remove('127.0.0.1')
        # should have no effect, only exact matches
        self.assertTrue(filter.is_allowed('127.0.0.1'))

        filter.whitelist_remove('127.0.0.0/24')
        self.assertFalse(filter.is_allowed('127.0.0.1'))

    def test_invalid_config_loading(self):
        filter = IPFilter()
        # mandatory keys not present
        with self.assertRaises(ValueError) as ctx:
            filter.load_config({'blacklist':[]})
        self.assertEqual("whitelist key not found", str(ctx.exception))

        # mandatory key blacklist not present
        with self.assertRaises(ValueError) as ctx:
            filter.load_config({'whitelist':[]})
        self.assertEqual("blacklist key not found", str(ctx.exception))

    def test_config_reset(self):
        filter = IPFilter()
        filter.blacklist_add('127.0.0.1')
        self.assertEqual(1, len(filter._config['blacklist']))
        filter.reset()
        self.assertEqual(filter.default_config,filter._config)



import asyncio
import struct
import weakref
import traceback
from typing import Optional
from neo3.network import node
from neo3.network.message import Message
from asyncio.streams import StreamReader, StreamReaderProtocol, StreamWriter
from asyncio import events
from neo3 import network_logger as logger


class NeoProtocol(StreamReaderProtocol):
    def __init__(self, *args, **kwargs):
        """

        Args:
            *args:
            **kwargs:
        """
        sr = StreamReader()
        self._stream_reader_orig = sr
        self._stream_reader_wr = weakref.ref(sr)
        self._stream_writer = None
        self.client = node.NeoNode(self)
        self._loop = events.get_event_loop()
        super().__init__(sr)

    def connection_made(self, transport: asyncio.transports.BaseTransport) -> None:
        super().connection_made(transport)
        self._stream_writer = StreamWriter(transport, self, self._stream_reader_orig, self._loop)

        if self.client:
            asyncio.create_task(self.client.connection_made(transport))

    def connection_lost(self, exc: Optional[Exception] = None) -> None:
        if self.client:
            task = asyncio.create_task(self.client.connection_lost(exc))
            task.add_done_callback(lambda args: super(NeoProtocol, self).connection_lost(exc))
        else:
            super().connection_lost(exc)

    async def send_message(self, message: Message) -> None:
        try:
            self._stream_writer.write(message.to_array())
            await self._stream_writer.drain()
        except ConnectionResetError:
            print("connection reset")
            self.connection_lost(ConnectionResetError())
        except ConnectionError:
            print("connection error")
            self.connection_lost(ConnectionError())
        except asyncio.CancelledError:
            print("task cancelled, closing connection")
            # mypy can't seem to deduce that CancelledError still derives from Exception
            self.connection_lost(asyncio.CancelledError())  # type: ignore
        except Exception as e:
            print(f"***** woah what happened here?! {traceback.format_exc()}")
            self.connection_lost(Exception())

    async def read_message(self, timeout: Optional[int] = 30) -> Optional[Message]:
        if timeout == 0:
            # avoid memleak. See: https://bugs.python.org/issue37042
            timeout = None

        async def _read():
            try:
                # readexactly can throw ConnectionResetError
                message_header = await self._stream_reader_orig.readexactly(3)
                payload_length = message_header[2]

                if payload_length == 0xFD:
                    len_bytes = await self._stream_reader_orig.readexactly(2)
                    payload_length, = struct.unpack("<H", len_bytes)
                elif payload_length == 0xFE:
                    len_bytes = await self._stream_reader_orig.readexactly(4)
                    payload_length, = struct.unpack("<I", len_bytes)
                elif payload_length == 0xFE:
                    len_bytes = await self._stream_reader_orig.readexactly(8)
                    payload_length, = struct.unpack("<Q", len_bytes)
                else:
                    len_bytes = b''

                if payload_length > Message.PAYLOAD_MAX_SIZE:
                    raise ValueError("Invalid format")

                payload_data = await self._stream_reader_orig.readexactly(payload_length)
                raw = message_header + len_bytes + payload_data

                try:
                    return Message.deserialize_from_bytes(raw)
                except Exception:
                    logger.debug(f"Failed to deserialize message: {traceback.format_exc()}")
                    return None

            except (ConnectionResetError, ValueError) as e:
                # ensures we break out of the main run() loop of Node, which triggers a disconnect callback to clean up
                self.client.disconnecting = True
                logger.debug(f"Failed to read message data for reason: {traceback.format_exc()}")
                return None
            except (asyncio.CancelledError, asyncio.IncompleteReadError):
                return None
            except Exception:
                # ensures we break out of the main run() loop of Node, which triggers a disconnect callback to clean up
                logger.debug(f"error read message 1 {traceback.format_exc()}")
                return None
        try:
            # logger.debug("trying to read message")
            return await asyncio.wait_for(_read(), timeout)
        except (asyncio.TimeoutError, asyncio.CancelledError):
            return None
        except Exception:
            logger.debug("error read message 2")
            traceback.print_exc()
            return None

    def disconnect(self) -> None:
        if self._stream_writer:
            self._stream_writer.close()



# The class is still called DesignationContract
import unittest
from neo3 import settings, contracts, blockchain
from neo3.core import types, cryptography
from neo3.network import payloads
from tests.contracts.interop.utils import test_engine


class TestDesignationContract(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # this triggers deployment of the native contracts
        blockchain.Blockchain()

        settings.network.standby_committee = ['02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765']
        settings.network.validators_count = 1

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()

    def test_assign_and_get_role(self):
        engine = test_engine(has_snapshot=True, has_container=True)
        # set signers list to our committee to pass check_committee() validation
        engine.script_container.signers = [payloads.Signer(
            types.UInt160.from_string("2006cf497676f551841a150550dc3f561c1b6c67"),
            payloads.WitnessScope.GLOBAL
        )]
        public_key1 = cryptography.KeyPair(b'\x01' * 32).public_key
        public_key2 = cryptography.KeyPair(b'\x02' * 32).public_key
        public_key3 = cryptography.KeyPair(b'\x03' * 32).public_key
        c = contracts.DesignationContract()
        c.designate_as_role(engine, contracts.DesignateRole.STATE_VALIDATOR, [public_key1, public_key2])
        c.designate_as_role(engine, contracts.DesignateRole.ORACLE, [public_key3])

        index = engine.snapshot.persisting_block.index + 1
        state_nodes = c.get_designated_by_role(engine.snapshot, contracts.DesignateRole.STATE_VALIDATOR, index)
        self.assertEqual(2, len(state_nodes))
        self.assertIn(public_key1, state_nodes)
        self.assertIn(public_key2, state_nodes)

        oracle_nodes = c.get_designated_by_role(engine.snapshot, contracts.DesignateRole.ORACLE, index)
        self.assertEqual(1, len(oracle_nodes))
        self.assertEqual(public_key3, oracle_nodes[0])



from __future__ import annotations
from enum import IntEnum
from neo3.core import Size as s, serialization
from typing import TypeVar

ConsensusMessage_t = TypeVar('ConsensusMessage_t', bound='ConsensusMessage')


class ConsensusMessageType(IntEnum):
    CHANGE_VIEW = 0x00,
    PREPARE_REQUEST = 0x20,
    PREPARE_RESPONSE = 0x21,
    COMMIT = 0x30,
    RECOVERY_REQUEST = 0x40,
    RECOVERY_MESSAGE = 0x41


class ConsensusMessage(serialization.ISerializable):
    """
    Base class for the various consensus messages
    """
    def __init__(self, type: ConsensusMessageType):
        self.type = type
        self.view_number: int = 0

    def __len__(self):
        return s.uint8 + s.uint8

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.type)
        writer.write_uint8(self.view_number)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.type = ConsensusMessageType(reader.read_uint8())
        self.view_number = reader.read_uint8()

    def deserialize_specialization_from_bytes(self, data: bytearray) -> ConsensusMessage_t:
        # TODO: not implemented. Requires all ConsensusMessage subclasses to be implemented. Low priority.
        pass

    @classmethod
    def _serializable_init(cls):
        return cls(ConsensusMessageType.CHANGE_VIEW)
